import hashlib
from pathlib import Path
from typing import Union
from warnings import warn

import numpy as np

from pypulseq import __version__
from pypulseq.supported_labels_rf_use import get_supported_labels, get_supported_rf_uses

version_major, version_minor, version_revision = __version__.split('.')[:3]


def write(self, file_name: Union[str, Path], create_signature, remove_duplicates=True) -> Union[str, None]:
    """
    Write the sequence data to the given filename using the open file format for MR sequences.

    See also `pypulseq.Sequence.read_seq.read()`.

    Parameters
    ----------
    file_name : str or Path
        File name of `.seq` file to be written to disk.
    create_signature : bool
    remove_duplicates : bool
        Before writing, remove and remap events that would be duplicates after
        the rounding done during writing

    Returns
    -------
    md5 or None : If create_signature is True, it returns the written .seq file's signature as a string,
    otherwise it returns None. Note that, if remove_duplicates is True, signature belongs to the
    deduplicated sequences signature, and not the Sequence that is stored in the Sequence object.

    Raises
    ------
    RuntimeError
        If an unsupported definition is encountered.
    """
    # `>.0f` for decimals.
    # `>g` to truncate insignificant zeros.
    file_name = Path(file_name)
    if file_name.suffix != '.seq':
        # Append .seq suffix
        file_name = file_name.with_suffix(file_name.suffix + '.seq')

    # If removing duplicates, make a copy of the sequence with the duplicate
    # events removed.
    if remove_duplicates:
        self = self.remove_duplicates()

    with open(file_name, 'w') as output_file:
        output_file.write('# Pulseq sequence file\n')
        output_file.write('# Created by PyPulseq\n\n')

        output_file.write('[VERSION]\n')
        output_file.write(f'major {self.version_major}\n')
        output_file.write(f'minor {self.version_minor}\n')
        output_file.write(f'revision {self.version_revision}\n')
        output_file.write('\n')

        if len(self.definitions) != 0:
            output_file.write('[DEFINITIONS]\n')
            keys = sorted(self.definitions.keys())
            values = [self.definitions[k] for k in keys]
            for block_counter in range(len(keys)):
                output_file.write(f'{keys[block_counter]} ')
                if isinstance(values[block_counter], str):
                    output_file.write(values[block_counter] + ' ')
                elif isinstance(values[block_counter], (int, float)):
                    output_file.write(f'{values[block_counter]:0.9g} ')
                elif isinstance(values[block_counter], (list, tuple, np.ndarray)):  # e.g. [FOV_x, FOV_y, FOV_z]
                    for i in range(len(values[block_counter])):
                        if isinstance(values[block_counter][i], (int, float)):
                            output_file.write(f'{values[block_counter][i]:0.9g} ')
                        else:
                            output_file.write(f'{values[block_counter][i]} ')
                else:
                    raise RuntimeError('Unsupported definition')
                output_file.write('\n')
            output_file.write('\n')

        output_file.write('# Format of blocks:\n')
        output_file.write('# NUM DUR RF  GX  GY  GZ  ADC  EXT\n')
        output_file.write('[BLOCKS]\n')
        id_format_width = '{:' + str(len(str(len(self.block_events)))) + 'd}'
        id_format_str = id_format_width + ' {:3d} {:3d} {:3d} {:3d} {:3d} {:2d} {:2d}\n'
        for block_counter in self.block_events:
            block_duration = self.block_durations[block_counter] / self.block_duration_raster
            block_duration_rounded = round(block_duration)

            assert abs(block_duration_rounded - block_duration) < 1e-6

            s = id_format_str.format(
                *(
                    block_counter,
                    block_duration_rounded,
                    *self.block_events[block_counter][1:],
                )
            )
            output_file.write(s)
        output_file.write('\n')

        if len(self.rf_library.data) != 0:
            output_file.write('# Format of RF events:\n')
            output_file.write('# id ampl. mag_id phase_id time_shape_id center delay freqPPm phasePPM freq phase use\n')
            output_file.write('# ..   Hz      ..       ..            ..     us    us     ppm  rad/MHz   Hz   rad  ..\n')
            output_file.write(f'# Field "use" is the initial of: {" ".join(get_supported_rf_uses()).strip()}\n')
            output_file.write('[RF]\n')
            id_format_str = (
                '{:.0f} {:12g} {:.0f} {:.0f} {:.0f} {:g} {:g} {:g} {:g} {:g} {:g} {:s}\n'  # Refer lines 20-21
            )
            for k in self.rf_library.data:
                lib_data1 = self.rf_library.data[k][0:4]
                lib_data2 = self.rf_library.data[k][6:10]
                center = self.rf_library.data[k][4] * 1e6  # us
                delay = round(self.rf_library.data[k][5] / self.rf_raster_time) * self.rf_raster_time * 1e6
                s = id_format_str.format(k, *lib_data1, center, delay, *lib_data2, self.rf_library.type[k])
                output_file.write(s)
            output_file.write('\n')

        grad_lib_values = np.array(list(self.grad_library.type.values()))
        arb_grad_mask = grad_lib_values == 'g' if self.grad_library.type else False
        trap_grad_mask = grad_lib_values == 't' if self.grad_library.type else False

        if np.any(arb_grad_mask):
            output_file.write('# Format of arbitrary gradients:\n')
            output_file.write(
                '#   time_shape_id of 0 means default timing (stepping with grad_raster starting at 1/2 of grad_raster)\n'
            )
            output_file.write('# id amplitude first last amp_shape_id time_shape_id delay\n')
            output_file.write('# ..      Hz/m  Hz/m Hz/m        ..         ..          us\n')
            output_file.write('[GRADIENTS]\n')
            id_format_str = '{:.0f} {:12g} {:12g} {:12g} {:.0f} {:.0f} {:.0f}\n'  # Refer lines 20-21
            keys = np.array(list(self.grad_library.data.keys()))
            for k in keys[arb_grad_mask]:
                s = id_format_str.format(
                    k,
                    *self.grad_library.data[k][:5],
                    round(self.grad_library.data[k][5] * 1e6),
                )
                output_file.write(s)
            output_file.write('\n')

        if np.any(trap_grad_mask):
            output_file.write('# Format of trapezoid gradients:\n')
            output_file.write('# id amplitude rise flat fall delay\n')
            output_file.write('# ..      Hz/m   us   us   us    us\n')
            output_file.write('[TRAP]\n')
            keys = np.array(list(self.grad_library.data.keys()))
            id_format_str = '{:2.0f} {:12g} {:3.0f} {:4.0f} {:3.0f} {:3.0f}\n'
            for k in keys[trap_grad_mask]:
                data = np.copy(self.grad_library.data[k])  # Make a copy to leave the original untouched
                data[1:] = np.round(1e6 * data[1:])
                """
                Python & Numpy always round to nearest even value - inconsistent with MATLAB Pulseq's .seq files.
                [1] https://stackoverflow.com/questions/29671945/format-string-rounding-inconsistent
                [2] https://stackoverflow.com/questions/50374779/how-to-avoid-incorrect-rounding-with-numpy-round
                """
                s = id_format_str.format(k, *data)
                output_file.write(s)
            output_file.write('\n')

        if len(self.adc_library.data) != 0:
            output_file.write('# Format of ADC events:\n')
            output_file.write('# id num dwell delay freqPPM phasePPM freq phase phase_id\n')
            output_file.write('# ..  ..    ns    us     ppm  rad/MHz   Hz   rad       ..\n')
            output_file.write('[ADC]\n')
            id_format_str = '{:.0f} {:.0f} {:.0f} {:.0f} {:g} {:g} {:g} {:g} {:.0f}\n'  # Refer lines 20-21
            for k in self.adc_library.data:
                data = np.multiply(self.adc_library.data[k][0:8], [1, 1e9, 1e6, 1, 1, 1, 1, 1])
                s = id_format_str.format(k, *data)
                output_file.write(s)
            output_file.write('\n')

        if len(self.extensions_library.data) != 0:
            output_file.write('# Format of extension lists:\n')
            output_file.write('# id type ref next_id\n')
            output_file.write('# next_id of 0 terminates the list\n')
            output_file.write('# Extension list is followed by extension specifications\n')
            output_file.write('[EXTENSIONS]\n')
            id_format_str = '{:.0f} {:.0f} {:.0f} {:.0f}\n'  # Refer lines 20-21
            for k in self.extensions_library.data:
                s = id_format_str.format(k, *np.round(self.extensions_library.data[k]))
                output_file.write(s)
            output_file.write('\n')

        if len(self.trigger_library.data) != 0:
            output_file.write('# Extension specification for digital output and input triggers:\n')
            output_file.write('# id type channel delay (us) duration (us)\n')
            output_file.write(f'extension TRIGGERS {self.get_extension_type_ID("TRIGGERS")}\n')
            id_format_str = '{:.0f} {:.0f} {:.0f} {:.0f} {:.0f}\n'  # Refer lines 20-21
            for k in self.trigger_library.data:
                s = id_format_str.format(k, *np.round(self.trigger_library.data[k] * np.array([1, 1, 1e6, 1e6])))
                output_file.write(s)
            output_file.write('\n')

        if len(self.label_set_library.data) != 0:
            labels = get_supported_labels()

            output_file.write('# Extension specification for setting labels:\n')
            output_file.write('# id set labelstring\n')
            tid = self.get_extension_type_ID('LABELSET')
            output_file.write(f'extension LABELSET {tid}\n')
            id_format_str = '{:.0f} {:.0f} {}\n'  # Refer lines 20-21
            for k in self.label_set_library.data:
                value = self.label_set_library.data[k][0]
                label_id = labels[int(self.label_set_library.data[k][1]) - 1]  # label_id is +1 in add_block()
                s = id_format_str.format(k, value, label_id)
                output_file.write(s)
            output_file.write('\n')

        if len(self.label_inc_library.data) != 0:
            labels = get_supported_labels()

            output_file.write('# Extension specification for setting labels:\n')
            output_file.write('# id set labelstring\n')
            tid = self.get_extension_type_ID('LABELINC')
            output_file.write(f'extension LABELINC {tid}\n')
            id_format_str = '{:.0f} {:.0f} {}\n'  # See comment at the beginning of this method definition
            for k in self.label_inc_library.data:
                value = self.label_inc_library.data[k][0]
                label_id = labels[self.label_inc_library.data[k][1] - 1]  # label_id is +1 in add_block()
                s = id_format_str.format(k, value, label_id)
                output_file.write(s)
            output_file.write('\n')

        if len(self.soft_delay_library.data) != 0:
            output_file.write('# Extension specification for soft delays:\n')
            output_file.write('# id num offset factor hint\n')
            output_file.write('# ..  ..     us     ..   ..\n')

            tid = self.get_extension_type_ID('DELAYS')
            output_file.write(f'extension DELAYS {tid}\n')
            id_format_str = '{:.0f} {:.0f} {:.0f} {:.0f} {}\n'

            for k in self.soft_delay_library.data:
                data = self.soft_delay_library.data[k]
                s = id_format_str.format(k, data[0], np.round(data[1] * 1e6), data[2], data[3])
                output_file.write(s)
            output_file.write('\n')

        if len(self.shape_library.data) != 0:
            output_file.write('# Sequence Shapes\n')
            output_file.write('[SHAPES]\n\n')
            for k in self.shape_library.data:
                shape_data = self.shape_library.data[k]
                s = 'shape_id {:.0f}\n'.format(k)
                output_file.write(s)
                s = 'num_samples {:.0f}\n'.format(shape_data[0])
                output_file.write(s)
                s = ('{:.9g}\n' * len(shape_data[1:])).format(*shape_data[1:])
                output_file.write(s)
                output_file.write('\n')

    if create_signature:  # Sign the file
        # Calculate digest
        with open(file_name, 'r') as output_file:
            buffer = output_file.read()

            md5 = hashlib.md5(buffer.encode('utf-8')).hexdigest()

        # Write signature
        with open(file_name, 'a') as output_file:
            output_file.write('\n[SIGNATURE]\n')
            output_file.write(
                '# This is the hash of the Pulseq file, calculated right before the [SIGNATURE] section was added\n'
            )
            output_file.write(
                '# It can be reproduced/verified with md5sum if the file trimmed to the position right above [SIGNATURE]\n'
            )
            output_file.write(
                '# The new line character preceding [SIGNATURE] BELONGS to the signature (and needs to be stripped away for '
                'recalculating/verification)\n'
            )
            output_file.write('Type md5\n')
            output_file.write(f'Hash {md5}\n')

        return md5


def write_v141(self, file_name: Union[str, Path], create_signature, remove_duplicates=True) -> Union[str, None]:
    """
    Write the sequence data to the given filename using the open file format for MR sequences.

    See also `pypulseq.Sequence.read_seq.read()`.

    Parameters
    ----------
    file_name : str or Path
        File name of `.seq` file to be written to disk.
    create_signature : bool
    remove_duplicates : bool
        Before writing, remove and remap events that would be duplicates after
        the rounding done during writing

    Returns
    -------
    md5 or None : If create_signature is True, it returns the written .seq file's signature as a string,
    otherwise it returns None. Note that, if remove_duplicates is True, signature belongs to the
    deduplicated sequences signature, and not the Sequence that is stored in the Sequence object.

    Raises
    ------
    RuntimeError
        If an unsupported definition is encountered.
    """
    # `>.0f` for decimals.
    # `>g` to truncate insignificant zeros.
    file_name = Path(file_name)
    if file_name.suffix != '.seq':
        # Append .seq suffix
        file_name = file_name.with_suffix(file_name.suffix + '.seq')

    # If removing duplicates, make a copy of the sequence with the duplicate
    # events removed.
    if remove_duplicates:
        self = self.remove_duplicates()

    with open(file_name, 'w') as output_file:
        output_file.write('# Pulseq sequence file\n')
        output_file.write('# Created by PyPulseq\n\n')

        output_file.write('[VERSION]\n')
        output_file.write('major 1\n')
        output_file.write('minor 4\n')
        output_file.write('revision 1\n')
        output_file.write('\n')

        ppm_to_hz = 1e-6 * self.system.gamma * self.system.B0

        if len(self.definitions) != 0:
            output_file.write('[DEFINITIONS]\n')
            keys = sorted(self.definitions.keys())
            values = [self.definitions[k] for k in keys]
            for block_counter in range(len(keys)):
                output_file.write(f'{keys[block_counter]} ')
                if isinstance(values[block_counter], str):
                    output_file.write(values[block_counter] + ' ')
                elif isinstance(values[block_counter], (int, float)):
                    output_file.write(f'{values[block_counter]:0.9g} ')
                elif isinstance(values[block_counter], (list, tuple, np.ndarray)):  # e.g. [FOV_x, FOV_y, FOV_z]
                    for i in range(len(values[block_counter])):
                        if isinstance(values[block_counter][i], (int, float)):
                            output_file.write(f'{values[block_counter][i]:0.9g} ')
                        else:
                            output_file.write(f'{values[block_counter][i]} ')
                else:
                    raise RuntimeError('Unsupported definition')
                output_file.write('\n')
            output_file.write('\n')

        output_file.write('# Format of blocks:\n')
        output_file.write('# NUM DUR RF  GX  GY  GZ  ADC  EXT\n')
        output_file.write('[BLOCKS]\n')
        id_format_width = '{:' + str(len(str(len(self.block_events)))) + 'd}'
        id_format_str = id_format_width + ' {:3d} {:3d} {:3d} {:3d} {:3d} {:2d} {:2d}\n'
        for block_counter in self.block_events:
            block_duration = self.block_durations[block_counter] / self.block_duration_raster
            block_duration_rounded = round(block_duration)

            assert abs(block_duration_rounded - block_duration) < 1e-6

            s = id_format_str.format(
                *(
                    block_counter,
                    block_duration_rounded,
                    *self.block_events[block_counter][1:],
                )
            )
            output_file.write(s)
        output_file.write('\n')

        if len(self.rf_library.data) != 0:
            output_file.write('# Format of RF events:\n')
            output_file.write('# id amplitude mag_id phase_id time_shape_id delay freq phase\n')
            output_file.write('# ..        Hz   ....     ....          ....    us   Hz   rad\n')
            output_file.write('[RF]\n')
            id_format_str = '{:.0f} {:12g} {:.0f} {:.0f} {:.0f} {:g} {:g} {:g}\n'  # Refer lines 20-21
            for k in self.rf_library.data:
                lib = self.rf_library.data[k]
                data1 = lib[0:4]
                data2 = tuple(a + b * ppm_to_hz for a, b in zip(lib[8:10], lib[6:8]))
                delay = round(lib[5] / self.rf_raster_time) * self.rf_raster_time * 1e6
                s = id_format_str.format(k, *data1, delay, *data2)
                output_file.write(s)
            output_file.write('\n')

        grad_lib_values = np.array(list(self.grad_library.type.values()))
        arb_grad_mask = grad_lib_values == 'g' if self.grad_library.type else False
        trap_grad_mask = grad_lib_values == 't' if self.grad_library.type else False

        if np.any(arb_grad_mask):
            output_file.write('# Format of arbitrary gradients:\n')
            output_file.write(
                '#   time_shape_id of 0 means default timing (stepping with grad_raster starting at 1/2 of grad_raster)\n'
            )
            output_file.write('# id amplitude amp_shape_id time_shape_id delay\n')
            output_file.write('# ..      Hz/m       ..         ..          us\n')
            output_file.write('[GRADIENTS]\n')
            id_format_str = '{:.0f} {:12g} {:.0f} {:.0f} {:.0f}\n'  # Refer lines 20-21
            keys = np.array(list(self.grad_library.data.keys()))
            for k in keys[arb_grad_mask]:
                s = id_format_str.format(
                    k,
                    self.grad_library.data[k][0],
                    *self.grad_library.data[k][3:5],
                    round(self.grad_library.data[k][5] * 1e6),
                )
                output_file.write(s)
            output_file.write('\n')

        if np.any(trap_grad_mask):
            output_file.write('# Format of trapezoid gradients:\n')
            output_file.write('# id amplitude rise flat fall delay\n')
            output_file.write('# ..      Hz/m   us   us   us    us\n')
            output_file.write('[TRAP]\n')
            keys = np.array(list(self.grad_library.data.keys()))
            id_format_str = '{:2.0f} {:12g} {:3.0f} {:4.0f} {:3.0f} {:3.0f}\n'
            for k in keys[trap_grad_mask]:
                data = np.copy(self.grad_library.data[k])  # Make a copy to leave the original untouched
                data[1:] = np.round(1e6 * data[1:])
                """
                Python & Numpy always round to nearest even value - inconsistent with MATLAB Pulseq's .seq files.
                [1] https://stackoverflow.com/questions/29671945/format-string-rounding-inconsistent
                [2] https://stackoverflow.com/questions/50374779/how-to-avoid-incorrect-rounding-with-numpy-round
                """
                s = id_format_str.format(k, *data)
                output_file.write(s)
            output_file.write('\n')

        if len(self.adc_library.data) != 0:
            output_file.write('# Format of ADC events:\n')
            output_file.write('# id num dwell delay freq phase\n')
            output_file.write('# ..  ..    ns    us   Hz   rad\n')
            output_file.write('[ADC]\n')
            id_format_str = '{:.0f} {:.0f} {:.0f} {:.0f} {:g} {:g}\n'  # Refer lines 20-21
            for k in self.adc_library.data:
                lib = self.adc_library.data[k]
                data1 = (lib[0], lib[1] * 1e9, lib[2] * 1e6)
                data2 = tuple(a + b * ppm_to_hz for a, b in zip(lib[5:7], lib[3:5]))
                s = id_format_str.format(k, *data1, *data2)
                output_file.write(s)
            output_file.write('\n')

        if len(self.extensions_library.data) != 0:
            output_file.write('# Format of extension lists:\n')
            output_file.write('# id type ref next_id\n')
            output_file.write('# next_id of 0 terminates the list\n')
            output_file.write('# Extension list is followed by extension specifications\n')
            output_file.write('[EXTENSIONS]\n')
            id_format_str = '{:.0f} {:.0f} {:.0f} {:.0f}\n'  # Refer lines 20-21
            for k in self.extensions_library.data:
                s = id_format_str.format(k, *np.round(self.extensions_library.data[k]))
                output_file.write(s)
            output_file.write('\n')

        if len(self.trigger_library.data) != 0:
            output_file.write('# Extension specification for digital output and input triggers:\n')
            output_file.write('# id type channel delay (us) duration (us)\n')
            output_file.write(f'extension TRIGGERS {self.get_extension_type_ID("TRIGGERS")}\n')
            id_format_str = '{:.0f} {:.0f} {:.0f} {:.0f} {:.0f}\n'  # Refer lines 20-21
            for k in self.trigger_library.data:
                s = id_format_str.format(k, *np.round(self.trigger_library.data[k] * np.array([1, 1, 1e6, 1e6])))
                output_file.write(s)
            output_file.write('\n')

        if len(self.label_set_library.data) != 0:
            labels = get_supported_labels()

            output_file.write('# Extension specification for setting labels:\n')
            output_file.write('# id set labelstring\n')
            tid = self.get_extension_type_ID('LABELSET')
            output_file.write(f'extension LABELSET {tid}\n')
            id_format_str = '{:.0f} {:.0f} {}\n'  # Refer lines 20-21
            for k in self.label_set_library.data:
                value = self.label_set_library.data[k][0]
                label_id = labels[int(self.label_set_library.data[k][1]) - 1]  # label_id is +1 in add_block()
                s = id_format_str.format(k, value, label_id)
                output_file.write(s)
            output_file.write('\n')

        if len(self.label_inc_library.data) != 0:
            labels = get_supported_labels()

            output_file.write('# Extension specification for setting labels:\n')
            output_file.write('# id set labelstring\n')
            tid = self.get_extension_type_ID('LABELINC')
            output_file.write(f'extension LABELINC {tid}\n')
            id_format_str = '{:.0f} {:.0f} {}\n'  # See comment at the beginning of this method definition
            for k in self.label_inc_library.data:
                value = self.label_inc_library.data[k][0]
                label_id = labels[self.label_inc_library.data[k][1] - 1]  # label_id is +1 in add_block()
                s = id_format_str.format(k, value, label_id)
                output_file.write(s)
            output_file.write('\n')

        if len(self.soft_delay_library.data) != 0:
            warn(
                'WARNING! The sequence in memory uses "soft delay" extension, which is incompatible with the file format v1.4.1. The produced Pulseq file is only partially valid and may fail to load or operate in some cases.'
            )

        if len(self.shape_library.data) != 0:
            output_file.write('# Sequence Shapes\n')
            output_file.write('[SHAPES]\n\n')
            for k in self.shape_library.data:
                shape_data = self.shape_library.data[k]
                s = 'shape_id {:.0f}\n'.format(k)
                output_file.write(s)
                s = 'num_samples {:.0f}\n'.format(shape_data[0])
                output_file.write(s)
                s = ('{:.9g}\n' * len(shape_data[1:])).format(*shape_data[1:])
                output_file.write(s)
                output_file.write('\n')

    if create_signature:  # Sign the file
        # Calculate digest
        with open(file_name, 'r') as output_file:
            buffer = output_file.read()

            md5 = hashlib.md5(buffer.encode('utf-8')).hexdigest()

        # Write signature
        with open(file_name, 'a') as output_file:
            output_file.write('\n[SIGNATURE]\n')
            output_file.write(
                '# This is the hash of the Pulseq file, calculated right before the [SIGNATURE] section was added\n'
            )
            output_file.write(
                '# It can be reproduced/verified with md5sum if the file trimmed to the position right above [SIGNATURE]\n'
            )
            output_file.write(
                '# The new line character preceding [SIGNATURE] BELONGS to the signature (and needs to be stripped away for '
                'recalculating/verification)\n'
            )
            output_file.write('Type md5\n')
            output_file.write(f'Hash {md5}\n')

        return md5



from types import SimpleNamespace
from typing import List

import numpy as np


def asc_to_acoustic_resonances(asc: dict) -> List[dict]:
    """
    Convert ASC dictionary from readasc to list of acoustic resonances

    Parameters
    ----------
    asc : dict
        ASC dictionary, see readasc

    Returns
    -------
    List[dict]
        List of acoustic resonances (specified by frequency and bandwidth fields).
    """
    if 'aflGCAcousticResonanceFrequency' in asc:
        freqs = asc['aflGCAcousticResonanceFrequency']
        bw = asc['aflGCAcousticResonanceBandwidth']
    else:
        freqs = asc['asGPAParameters'][0]['sGCParameters']['aflAcousticResonanceFrequency']
        bw = asc['asGPAParameters'][0]['sGCParameters']['aflAcousticResonanceBandwidth']

    return [{'frequency': f, 'bandwidth': b} for f, b in zip(freqs.values(), bw.values(), strict=False) if f != 0]


def asc_to_hw(asc: dict, cardiac_model: bool = False) -> SimpleNamespace:
    """
    Convert ASC dictionary from readasc to SAFE hardware description.

    Parameters
    ----------
    asc : dict
        ASC dictionary, see readasc
    cardiac_model : bool
        Whether or not to read the cardiac stimulation model instead of the
        default PNS model (returns None if not available)

    Returns
    -------
    SimpleNamespace
        SAFE hardware description
    """
    hw = SimpleNamespace()

    if 'asCOMP' in asc and 'tName' in asc['asCOMP']:
        hw.name = asc['asCOMP']['tName']
    else:
        hw.name = 'unknown'

    if 'GradPatSup' in asc:
        asc_pns = asc['GradPatSup']['Phys']['PNS']
    else:
        asc_pns = asc

    if cardiac_model:
        if 'GradPatSup' in asc and 'CarNS' in asc['GradPatSup']['Phys']:
            asc_pns = asc['GradPatSup']['Phys']['CarNS']
        else:
            return None

    hw.x = SimpleNamespace()
    hw.x.tau1 = asc_pns['flGSWDTauX'][0]  # ms
    hw.x.tau2 = asc_pns['flGSWDTauX'][1]  # ms
    hw.x.tau3 = asc_pns['flGSWDTauX'][2]  # ms
    hw.x.a1 = asc_pns['flGSWDAX'][0]
    hw.x.a2 = asc_pns['flGSWDAX'][1]
    hw.x.a3 = asc_pns['flGSWDAX'][2]
    hw.x.stim_limit = asc_pns['flGSWDStimulationLimitX']  # T/m/s
    hw.x.stim_thresh = asc_pns['flGSWDStimulationThresholdX']  # T/m/s

    hw.y = SimpleNamespace()
    hw.y.tau1 = asc_pns['flGSWDTauY'][0]  # ms
    hw.y.tau2 = asc_pns['flGSWDTauY'][1]  # ms
    hw.y.tau3 = asc_pns['flGSWDTauY'][2]  # ms
    hw.y.a1 = asc_pns['flGSWDAY'][0]
    hw.y.a2 = asc_pns['flGSWDAY'][1]
    hw.y.a3 = asc_pns['flGSWDAY'][2]
    hw.y.stim_limit = asc_pns['flGSWDStimulationLimitY']  # T/m/s
    hw.y.stim_thresh = asc_pns['flGSWDStimulationThresholdY']  # T/m/s

    hw.z = SimpleNamespace()
    hw.z.tau1 = asc_pns['flGSWDTauZ'][0]  # ms
    hw.z.tau2 = asc_pns['flGSWDTauZ'][1]  # ms
    hw.z.tau3 = asc_pns['flGSWDTauZ'][2]  # ms
    hw.z.a1 = asc_pns['flGSWDAZ'][0]
    hw.z.a2 = asc_pns['flGSWDAZ'][1]
    hw.z.a3 = asc_pns['flGSWDAZ'][2]
    hw.z.stim_limit = asc_pns['flGSWDStimulationLimitZ']  # T/m/s
    hw.z.stim_thresh = asc_pns['flGSWDStimulationThresholdZ']  # T/m/s

    if 'asGPAParameters' in asc:
        hw.x.g_scale = asc['asGPAParameters'][0]['sGCParameters']['flGScaleFactorX']
        hw.y.g_scale = asc['asGPAParameters'][0]['sGCParameters']['flGScaleFactorY']
        hw.z.g_scale = asc['asGPAParameters'][0]['sGCParameters']['flGScaleFactorZ']
    else:
        print('Warning: Gradient scale factors not in ASC file: assuming 1/pi')
        hw.x.g_scale = 1 / np.pi
        hw.y.g_scale = 1 / np.pi
        hw.z.g_scale = 1 / np.pi

    return hw



from types import SimpleNamespace
from typing import Tuple, Union

try:
    from typing import Self
except ImportError:
    from typing import TypeVar

    Self = TypeVar('Self', bound='EventLibrary')

import math

import numpy as np


class EventLibrary:
    """
    Defines an event library to maintain a list of events. Provides methods to insert new data and find existing data.

    Sequence Properties:
    - data - A struct array with field 'array' to store data of varying lengths, remaining compatible with codegen.
    - type - Type to distinguish events in the same class (e.g. trapezoids and arbitrary gradients)

    Sequence Methods:
    - find - Find an event in the library
    - insert - Add a new event to the library

    See also `Sequence.py`.

    Attributes
    ----------
    data : dict{str: numpy.array}
        Key-value pairs of event keys and corresponding data.
    type : dict{str, str}
        Key-value pairs of event keys and corresponding event types.
    keymap : dict{str, int}
        Key-value pairs of data values and corresponding event keys.
    """

    def __init__(self, numpy_data=False):
        self.data = {}
        self.type = {}
        self.keymap = {}
        self.next_free_ID = 1
        self.numpy_data = numpy_data

    def __str__(self) -> str:
        s = 'EventLibrary:'
        s += '\ndata: ' + str(len(self.data))
        s += '\ntype: ' + str(len(self.type))
        return s

    def find(self, new_data: np.ndarray) -> Tuple[int, bool]:
        """
        Finds data `new_data` in event library.

        Parameters
        ----------
        new_data : numpy.ndarray
            Data to be found in event library.

        Returns
        -------
        key_id : int
            Key of `new_data` in event library, if found.
        found : bool
            If `new_data` was found in the event library or not.
        """
        if self.numpy_data:
            new_data = np.asarray(new_data)
            key = new_data.tobytes()
        else:
            key = tuple(new_data)

        if key in self.keymap:
            key_id = self.keymap[key]
            found = True
        else:
            key_id = self.next_free_ID
            found = False

        return key_id, found

    def find_or_insert(self, new_data: np.ndarray, data_type: str = str()) -> Tuple[int, bool]:
        """
        Lookup a data structure in the given library and return the index of the data in the library. If the data does
        not exist in the library it is inserted right away. The data is a 1xN array with event-specific data.

        See also  insert `pypulseq.Sequence.sequence.Sequence.add_block()`.

        Parameters
        ----------
        new_data : numpy.ndarray
            Data to be found (or added, if not found) in event library.
        data_type : str, default=str()
            Type of data.

        Returns
        -------
        key_id : int
            Key of `new_data` in event library, if found.
        found : bool
            If `new_data` was found in the event library or not.
        """
        if self.numpy_data:
            new_data = np.asarray(new_data)
            new_data.flags.writeable = False
            key = new_data.tobytes()
        else:
            key = tuple(new_data)

        if key in self.keymap:
            key_id = self.keymap[key]
            found = True
        else:
            key_id = self.next_free_ID
            found = False

            # Insert
            self.data[key_id] = new_data

            if data_type != str():
                self.type[key_id] = data_type

            self.keymap[key] = key_id
            self.next_free_ID = key_id + 1  # Update next_free_id

        return key_id, found

    def insert(self, key_id: int, new_data: np.ndarray | list, data_type: str = str()) -> int:
        """
        Add event to library.

        See also `pypulseq.event_library.EventLibrary.find()`.

        Parameters
        ----------
        key_id : int
            Key of `new_data`.
        new_data : numpy.ndarray or list
            Data to be inserted into event library.
        data_type : str, default=str()
            Data type of `new_data`.

        Returns
        -------
        key_id : int
            Key ID of inserted event.
        """
        if isinstance(key_id, float):
            key_id = int(key_id)

        if key_id == 0:
            key_id = self.next_free_ID

        if self.numpy_data:
            new_data = np.asarray(new_data)
            new_data.flags.writeable = False
            key = new_data.tobytes()
        else:
            key = tuple(new_data)

        self.data[key_id] = new_data
        if data_type != str():
            self.type[key_id] = data_type

        self.keymap[key] = key_id

        if key_id >= self.next_free_ID:
            self.next_free_ID = key_id + 1  # Update next_free_id

        return key_id

    def get(self, key_id: int) -> dict:
        """

        Parameters
        ----------
        key_id : int

        Returns
        -------
        dict
        """
        return {
            'key': key_id,
            'data': self.data[key_id],
            'type': self.type[key_id],
        }

    def out(self, key_id: int) -> SimpleNamespace:
        """
        Get element from library by key.

        See also `pypulseq.event_library.EventLibrary.find()`.

        Parameters
        ----------
        key_id : int

        Returns
        -------
        out : SimpleNamespace
        """
        out = SimpleNamespace()
        out.key = key_id
        out.data = self.data[key_id]
        out.type = self.type[key_id]

        return out

    def update(
        self,
        key_id: int,
        old_data: Union[np.ndarray, None],  # noqa: ARG002
        new_data: np.ndarray,
        data_type: str = str(),
    ):
        """
        Parameters
        ----------
        key_id : int
        old_data : numpy.ndarray (Ignored!)
        new_data : numpy.ndarray
        data_type : str, default=str()
        """
        if key_id in self.data and self.data[key_id] in self.keymap:
            del self.keymap[self.data[key_id]]

        self.insert(key_id, new_data, data_type)

    def update_data(
        self,
        key_id: int,
        old_data: np.ndarray,
        new_data: np.ndarray,
        data_type: str = str(),
    ):
        """
        Parameters
        ----------
        key_id : int
        old_data : np.ndarray (Ignored!)
        new_data : np.ndarray
        data_type : str
        """
        self.update(key_id, old_data, new_data, data_type)

    def remove_duplicates(self, digits: Union[int, Tuple[int]]) -> Tuple[Self, dict]:
        """
        Remove duplicate events from this event library by rounding the data
        according to the significant `digits` specification, and then removing
        duplicate events.
        Returns a new event library, leaving the current one intact.

        Parameters
        ----------
        digits : Union[int, List[int]]
            For libraries with `numpy_data == True`:
                A single number specifying the number of significant digits
                after rounding.
            Otherwise:
                A tuple of numbers specifying the number of significant digits
                after rounding for each entry in the event data tuple.

        Returns
        -------
        new_library : EventLibrary
            Event library with the duplicate events removed
        mapping : dict
            Dictionary containing a mapping of IDs in the old library to IDs
            in the new library.
        """

        def round_data(data: Tuple[float], digits: Tuple[int]) -> Tuple[float]:
            """
            Round the data tuple to a specified number of significant digits,
            specified by `digits`. Rounding behavior is similar to the {.Ng}
            format specifier if N > 0, and similar to {.0f} otherwise.
            """
            return tuple(
                round(d, dig - math.ceil(math.log10(abs(d) + 1e-12)) if dig > 0 else -dig)
                for d, dig in zip(data, digits, strict=False)
            )

        def round_data_numpy(data: np.ndarray, digits: int) -> np.ndarray:
            """
            Round the data array to a specified number of significant digits,
            specified by `digits`. Rounding behavior is similar to the {.Ng}
            format specifier if N > 0, and similar to {.0f} otherwise.
            """
            mags = 10 ** (digits - (np.ceil(np.log10(abs(data) + 1e-12))) if digits > 0 else -digits)
            result = np.round(data * mags) / mags
            result.flags.writeable = False
            return result

        # Round library data based on `digits` specification
        if self.numpy_data:
            rounded_data = {x: round_data_numpy(self.data[x], digits) for x in self.data}
        else:
            rounded_data = {x: round_data(self.data[x], digits) for x in self.data}

        # Initialize filtered library
        new_library = EventLibrary(numpy_data=self.numpy_data)

        # Initialize ID mapping. Always include 0:0 to allow the mapping dict
        # to be used for mapping block_events (which can contain 0, i.e. no
        # event)
        mapping = {0: 0}

        # Recreate library using rounded values
        for k, v in sorted(rounded_data.items()):
            mapping[k], _ = new_library.find_or_insert(v, self.type[k] if k in self.type else str())

        return new_library, mapping



import importlib
import math
import os
from pathlib import Path
from types import SimpleNamespace
from unittest.mock import patch

import matplotlib.pyplot as plt
import numpy as np
import pypulseq as pp
import pytest
from _pytest.python_api import ApproxBase
from pypulseq import Sequence

expected_output_path = Path(__file__).parent / 'expected_output'


class Approx(ApproxBase):
    """
    Extension of pytest.approx that also handles approximate equality
    recursively within dicts, lists, tuples, and SimpleNamespace
    """

    def __repr__(self):
        return str(self.expected)

    def __eq__(self, actual):
        # if type(actual) != type(self.expected):
        #     return False
        if isinstance(self.expected, dict):
            if set(self.expected.keys()) != set(actual.keys()):
                return False

            for k in self.expected:
                if actual[k] != Approx(self.expected[k], rel=self.rel, abs=self.abs, nan_ok=self.nan_ok):
                    return False
            return True
        elif isinstance(self.expected, (list, tuple)):
            if len(self.expected) != len(actual):
                return False

            for e, a in zip(self.expected, actual, strict=False):
                if a != Approx(e, rel=self.rel, abs=self.abs, nan_ok=self.nan_ok):
                    return False
            return True
        elif isinstance(self.expected, SimpleNamespace):
            return actual.__dict__ == Approx(self.expected.__dict__, rel=self.rel, abs=self.abs, nan_ok=self.nan_ok)
        else:
            return actual == pytest.approx(self.expected, rel=self.rel, abs=self.abs, nan_ok=self.nan_ok)

    def _repr_compare(self, actual):
        # if type(actual) != type(self.expected):
        #     return [f'Actual and expected types do not match: {type(actual)} != {type(self.expected)}']
        if isinstance(self.expected, dict):
            if set(self.expected.keys()) != set(actual.keys()):
                return [f'Actual and expected keys do not match: {set(actual.keys())} != {set(self.expected.keys())}']

            r = []
            for k in self.expected:
                approx_obj = Approx(self.expected[k], rel=self.rel, abs=self.abs, nan_ok=self.nan_ok)
                if actual[k] != approx_obj:
                    r += [f'{k} does not match:']
                    r += [f'  {x}' for x in approx_obj._repr_compare(actual[k])]
            return r
        elif isinstance(self.expected, (list, tuple)):
            if len(self.expected) != len(actual):
                return [f'Actual and expected lengths do not match: {len(actual)} != {len(self.expected)}']
            r = []
            for i, (e, a) in enumerate(zip(self.expected, actual, strict=False)):
                approx_obj = Approx(e, rel=self.rel, abs=self.abs, nan_ok=self.nan_ok)
                if a != approx_obj:
                    r += [f'Index {i} does not match:']
                    r += [f'  {x}' for x in approx_obj._repr_compare(a)]
            return r
        elif isinstance(self.expected, SimpleNamespace):
            return Approx(self.expected.__dict__, rel=self.rel, abs=self.abs, nan_ok=self.nan_ok)._repr_compare(
                actual.__dict__
            )
        else:
            return pytest.approx(self.expected, rel=self.rel, abs=self.abs, nan_ok=self.nan_ok)._repr_compare(actual)


# Dummy sequence which contains only gaussian pulses with different parameters
def seq_make_gauss_pulses():
    seq = Sequence()
    seq.add_block(pp.make_gauss_pulse(flip_angle=1))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_gauss_pulse(flip_angle=1, delay=1e-3))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_gauss_pulse(flip_angle=math.pi / 2))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_gauss_pulse(flip_angle=math.pi / 2, duration=1e-3))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_gauss_pulse(flip_angle=math.pi / 2, duration=2e-3, phase_offset=math.pi / 2))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_gauss_pulse(flip_angle=math.pi / 2, duration=1e-3, phase_offset=math.pi / 2, freq_offset=1e3))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_gauss_pulse(flip_angle=math.pi / 2, duration=1e-3, time_bw_product=1))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_gauss_pulse(flip_angle=math.pi / 2, duration=1e-3, apodization=0.1))

    return seq


# Dummy sequence which contains only sinc pulses with different parameters
def seq_make_sinc_pulses():
    seq = Sequence()
    seq.add_block(pp.make_sinc_pulse(flip_angle=1))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_sinc_pulse(flip_angle=1, delay=1e-3))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_sinc_pulse(flip_angle=math.pi / 2))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_sinc_pulse(flip_angle=math.pi / 2, duration=1e-3))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_sinc_pulse(flip_angle=math.pi / 2, duration=2e-3, phase_offset=math.pi / 2))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_sinc_pulse(flip_angle=math.pi / 2, duration=1e-3, phase_offset=math.pi / 2, freq_offset=1e3))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_sinc_pulse(flip_angle=math.pi / 2, duration=1e-3, time_bw_product=1))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_sinc_pulse(flip_angle=math.pi / 2, duration=1e-3, apodization=0.1))

    return seq


# Dummy sequence which contains only block pulses with different parameters
def seq_make_block_pulses():
    seq = Sequence()
    seq.add_block(pp.make_block_pulse(flip_angle=1, duration=4e-3))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_block_pulse(flip_angle=1, delay=1e-3, duration=4e-3))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_block_pulse(flip_angle=math.pi / 2, duration=4e-3))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_block_pulse(flip_angle=math.pi / 2, duration=1e-3))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_block_pulse(flip_angle=math.pi / 2, duration=2e-3, phase_offset=math.pi / 2))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_block_pulse(flip_angle=math.pi / 2, duration=1e-3, phase_offset=math.pi / 2, freq_offset=1e3))
    seq.add_block(pp.make_delay(1))
    seq.add_block(pp.make_block_pulse(flip_angle=math.pi / 2, duration=1e-3, time_bw_product=1))

    return seq


# Basic sequence with gradients in all channels, some which are identical after rounding.
def seq1():
    seq = Sequence()
    seq.add_block(pp.make_block_pulse(math.pi / 4, duration=1e-3))
    seq.add_block(pp.make_trapezoid('x', area=1000))
    seq.add_block(pp.make_trapezoid('y', area=-500.00001))
    seq.add_block(pp.make_trapezoid('z', area=100))
    seq.add_block(pp.make_trapezoid('x', area=-1000), pp.make_trapezoid('y', area=500))
    seq.add_block(pp.make_trapezoid('y', area=-500), pp.make_trapezoid('z', area=1000))
    seq.add_block(pp.make_trapezoid('x', area=-1000), pp.make_trapezoid('z', area=1000.00001))

    return seq


# Basic spin-echo sequence structure
def seq2():
    seq = Sequence()
    seq.add_block(pp.make_block_pulse(math.pi / 2, duration=1e-3))
    seq.add_block(pp.make_trapezoid('x', area=1000))
    seq.add_block(pp.make_trapezoid('x', area=-1000))
    seq.add_block(pp.make_block_pulse(math.pi, duration=1e-3))
    seq.add_block(pp.make_trapezoid('x', area=-500))
    seq.add_block(pp.make_trapezoid('x', area=1000, duration=10e-3), pp.make_adc(num_samples=100, duration=10e-3))

    return seq


# Basic GRE sequence with INC labels
def seq3():
    seq = Sequence()

    for i in range(10):
        seq.add_block(pp.make_block_pulse(math.pi / 8, duration=1e-3))
        seq.add_block(pp.make_trapezoid('x', area=1000))
        seq.add_block(pp.make_trapezoid('y', area=-500 + i * 100))
        seq.add_block(pp.make_trapezoid('x', area=-500))
        seq.add_block(
            pp.make_trapezoid('x', area=1000, duration=10e-3),
            pp.make_adc(num_samples=100, duration=10e-3),
            pp.make_label(label='LIN', type='INC', value=1),
        )

    return seq


# Basic GRE sequence with SET labels
def seq4():
    seq = Sequence()

    for i in range(10):
        seq.add_block(pp.make_block_pulse(math.pi / 8, duration=1e-3))
        seq.add_block(pp.make_trapezoid('x', area=1000))
        seq.add_block(pp.make_trapezoid('y', area=-500 + i * 100))
        seq.add_block(pp.make_trapezoid('x', area=-500))
        seq.add_block(
            pp.make_trapezoid('x', area=1000, duration=10e-3),
            pp.make_adc(num_samples=100, duration=10e-3),
            pp.make_label(label='LIN', type='SET', value=i),
        )

    return seq


# GRE sequence with preceding noise acquisition and labels
def seq5():
    sys = pp.Opts()
    seq = Sequence(sys)
    rf, gz, gzr = pp.make_sinc_pulse(flip_angle=math.pi / 8, duration=1e-3, slice_thickness=3e-3, return_gz=True)
    gx = pp.make_trapezoid(channel='x', flat_area=32 * 1 / 0.3, flat_time=32 * 1e-4, system=sys)
    adc = pp.make_adc(num_samples=32, duration=gx.flat_time, delay=gx.rise_time, system=sys)
    gx_pre = pp.make_trapezoid(channel='x', area=-gx.area / 2, duration=1e-3, system=sys)
    phase_areas = -(np.arange(32) - 32 / 2) * (1 / 0.3)

    seq.add_block(pp.make_label(label='LIN', type='SET', value=0), pp.make_label(label='SLC', type='SET', value=0))
    seq.add_block(pp.make_adc(num_samples=1000, duration=1e-3), pp.make_label(label='NOISE', type='SET', value=True))
    seq.add_block(pp.make_label(label='NOISE', type='SET', value=False))
    seq.add_block(pp.make_delay(sys.rf_dead_time))

    for pe in range(32):
        gy_pre = pp.make_trapezoid(channel='y', area=phase_areas[pe], duration=1e-3, system=sys)

        seq.add_block(rf, gz)
        seq.add_block(gx_pre, gy_pre, gzr)
        seq.add_block(gx, adc, pp.make_label(label='LIN', type='SET', value=pe))

        gy_pre.amplitude = -gy_pre.amplitude
        seq.add_block(gx_pre, gy_pre, pp.make_delay(10e-3))

    return seq


# Basic GRE sequence with Soft Delay
def seq6():
    seq = Sequence()

    for i in range(10):
        seq.add_block(pp.make_block_pulse(math.pi / 8, duration=1e-3))
        seq.add_block(pp.make_trapezoid('x', area=1000))
        seq.add_block(pp.make_trapezoid('y', area=-500 + i * 100))
        seq.add_block(pp.make_trapezoid('x', area=-500))
        seq.add_block(pp.make_soft_delay(numID=0, hint='TE', offset=1, factor=1.0, default_duration=10e-6))
        seq.add_block(
            pp.make_trapezoid('x', area=1000, duration=10e-3),
            pp.make_adc(num_samples=100, duration=10e-3),
        )

    return seq


# List of all sequence functions that will be tested with the test functions below.
sequence_zoo = [seq_make_gauss_pulses, seq_make_sinc_pulses, seq_make_block_pulses, seq1, seq2, seq3, seq4, seq5, seq6]


# List of example sequences in pypulseq/seq_examples/scripts/ to add as
# sequence tests.
seq_examples = [
    'write_gre',
    'write_gre_label',
    'write_gre_label_softdelay',
    'write_haste',
    'write_radial_gre',
    'write_tse',
    'write_epi',
    'write_epi_label',
    'write_epi_se',
    'write_epi_se_rs',
    'write_mprage',
    'write_ute',
]

# Create a seq_func for each example script and add it to the list of sequences.
# Defining a new function ensures that pytest understands the name is
# e.g. `write_gre` instead of `main`.
# Derive the relative path from the test file to the examples folder
examples_dir = Path(__file__).resolve().parents[1] / 'examples' / 'scripts'


def make_test_func(example):
    def test_func(module_name=f'examples.{example}'):
        spec = importlib.util.spec_from_file_location(module_name, examples_dir / f'{example}.py')
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        return module.main()

    test_func.__name__ = example

    return test_func


for example in seq_examples:
    sequence_zoo.append(make_test_func(example))


# Main Sequence test class
# Note that pytest creates a new instance of the TestSequence class for each
# test. To prevent calling seq_func() over and over, we assign it as a class
# variable, and by using parametrize with scope='class' the order of the tests
# is per seq_func instead of per test.
@pytest.mark.parametrize('seq_func', sequence_zoo, scope='class')
class TestSequence:
    # Base test that just runs the sequence function and keeps the result
    # for the next tests.
    def test_sequence(self, seq_func):
        # Reset TestSequence.seq in case seq_func throws an exception (the
        # other tests will still run, but will result in AttributeErrors)
        TestSequence.seq = None
        TestSequence.seq = seq_func()

    # This "test" rewrites the expected .seq output files when SAVE_EXPECTED is
    # set in the environment variables.
    # E.g. in a unix-based system, run: SAVE_EXPECTED=1 pytest test_sequence.py
    @pytest.mark.skipif(not os.environ.get('SAVE_EXPECTED'), reason='Only save sequence files when requested')
    def test_save_expected(self, seq_func):
        seq_name = str(seq_func.__name__)
        TestSequence.seq.write(expected_output_path / (seq_name + '.seq'))

    # Test sequence.plot() method
    def test_plot(self, seq_func):
        if seq_func.__name__ in ['seq1', 'seq2', 'seq3', 'seq4', 'seq5']:
            with patch('matplotlib.pyplot.show'):
                TestSequence.seq.plot()
                TestSequence.seq.plot(show_blocks=True)
                TestSequence.seq.plot(time_range=(0, 1e-3))
                TestSequence.seq.plot(time_disp='ms')
                TestSequence.seq.plot(grad_disp='mT/m')
                plt.close('all')

    # Test sequence.test_report() method
    def test_test_report(self, seq_func):
        if seq_func.__name__ in seq_examples or seq_func.__name__ in ['seq2', 'seq3', 'seq4', 'seq5', 'seq6']:
            report = TestSequence.seq.test_report()
            assert isinstance(report, str), 'test_report() did not return a string'
            assert len(report) > 0, 'test_report() returned an empty string'

    # Test whether the sequence is the approximately the same after writing a .seq
    # file and reading it back in.
    def test_writeread(self, seq_func, tmp_path, compare_seq_file):
        seq_name = str(seq_func.__name__)
        output_filename = tmp_path / (seq_name + '.seq')

        seq = TestSequence.seq

        # Write sequence to file
        seq.write(output_filename)

        # Check if written sequence file matches expected sequence file
        compare_seq_file(output_filename, expected_output_path / (seq_name + '.seq'))

        # Read written sequence file back in
        seq2 = pp.Sequence(system=seq.system)
        seq2.read(output_filename)

        # Clean up written sequence file
        output_filename.unlink()

        # Test for approximate equality of all blocks
        assert list(seq2.block_events.keys()) == list(seq.block_events.keys()), 'Sequence block IDs are not identical'
        for block_counter in seq.block_events:
            block_orig = seq.get_block(block_counter)
            block_compare = seq2.get_block(block_counter)

            # if hasattr(block_orig, 'rf') and hasattr(block_orig.rf, 'use'):
            #     from copy import deepcopy

            #     block_orig = deepcopy(block_orig)
            #     block_orig.rf.use = 'undefined'

            assert block_compare == Approx(block_orig, abs=1e-5, rel=1e-5), f'Block {block_counter} does not match'

        # Test for approximate equality of all gradient waveforms
        for a, b, channel in zip(seq2.get_gradients(), seq.get_gradients(), ['x', 'y', 'z'], strict=False):
            if a is None and b is None:
                continue
            assert a is not None and b is not None

            # TODO: C[0] is slope of gradient, on the order of max_slew? So expect abs rounding errors in range of 1e2?
            assert a.x == Approx(b.x, abs=1e-5, rel=1e-5), (
                f'Time axis of gradient waveform for channel {channel} does not match'
            )
            assert a.c[0] == Approx(b.c[0], abs=1e2, rel=1e-3), (
                f'First-order coefficients of piecewise-polynomial gradient waveform for channel {channel} do not match'
            )
            assert a.c[1] == Approx(b.c[1], abs=1e-5, rel=1e-5), (
                f'Zero-order coefficients of piecewise-polynomial gradient waveform for channel {channel} do not match'
            )

        # Restore RF use for k-space calculation
        for block_counter in seq.block_events:
            block_orig = seq.get_block(block_counter)
            # if hasattr(block_orig, 'rf') and hasattr(block_orig.rf, 'use'):
            #     block_compare = seq2.get_block(block_counter)
            #     block_compare.rf.use = block_orig.rf.use

        # Test for approximate equality of kspace calculation
        assert seq2.calculate_kspace() == Approx(seq.calculate_kspace(), abs=1e-1, nan_ok=True)

        # Test whether labels are the same
        labels_seq = seq.evaluate_labels(evolution='blocks')
        labels_seq2 = seq2.evaluate_labels(evolution='blocks')

        assert labels_seq.keys() == labels_seq2.keys(), 'Sequences do not contain the same set of labels'

        for label in labels_seq:
            assert (labels_seq[label] == labels_seq2[label]).all(), f'Label {label} does not match'

    # Test whether the sequence is approximately the same after recreating it by
    # getting all blocks with get_block and inserting them into a new sequence
    # with add_block.
    # NOTE: In order to keep the order of shapes the same, sequences need to
    #       put RF events before gradient events and order arbitrary/extended
    #       gradient events in X, Y, Z order when passing them to
    #       seq.add_block(...). i.e. seq.add_block(rf, gx, gy, gz)
    def test_recreate(self, seq_func):  # noqa: ARG002
        seq = TestSequence.seq

        # Insert blocks from sequence into a new sequence
        seq2 = pp.Sequence(system=seq.system)
        for b in seq.block_events:
            seq2.add_block(seq.get_block(b))

        # Test for approximate equality of all blocks
        assert list(seq2.block_events.keys()) == list(seq.block_events.keys()), 'Sequence block IDs are not identical'
        for block_counter in seq.block_events:
            assert seq2.get_block(block_counter) == Approx(seq.get_block(block_counter), abs=1e-9, rel=1e-9), (
                f'Block {block_counter} does not match'
            )

        # Test for approximate equality of all gradient waveforms
        for a, b, channel in zip(seq2.get_gradients(), seq.get_gradients(), ['x', 'y', 'z'], strict=False):
            if a is None and b is None:
                continue
            assert a is not None and b is not None

            assert a.x == Approx(b.x, abs=1e-9, rel=1e-9), (
                f'Time axis of gradient waveform for channel {channel} does not match'
            )
            assert a.c[0] == Approx(b.c[0], abs=1e-9, rel=1e-9), (
                f'First-order coefficients of piecewise-polynomial gradient waveform for channel {channel} do not match'
            )
            assert a.c[1] == Approx(b.c[1], abs=1e-9, rel=1e-9), (
                f'Zero-order coefficients of piecewise-polynomial gradient waveform for channel {channel} do not match'
            )

        # Test for approximate equality of kspace calculation
        assert seq2.calculate_kspace() == Approx(seq.calculate_kspace(), abs=1e-6, nan_ok=True)

        # Test whether labels are the same
        labels_seq = seq.evaluate_labels(evolution='blocks')
        labels_seq2 = seq2.evaluate_labels(evolution='blocks')

        assert labels_seq.keys() == labels_seq2.keys(), 'Sequences do not contain the same set of labels'

        for label in labels_seq:
            assert (labels_seq[label] == labels_seq2[label]).all(), f'Label {label} does not match'



from types import SimpleNamespace
from typing import Tuple

import numpy as np


def calc_rf_center(rf: SimpleNamespace) -> Tuple[float, float]:
    """
    Calculate the time point of the effective rotation calculated as the peak of the radio-frequency amplitude for the
    shaped pulses and the center of the pulse for the block pulses. Zero padding in the radio-frequency pulse is
    considered as a part of the shape. Delay field of the radio-frequency object is not taken into account.

    Parameters
    ----------
    rf : SimpleNamespace
        Radio-frequency pulse event.

    Returns
    -------
    time_center : float
        Time point of the center of the radio-frequency pulse.
    id_center : float
        Corresponding position of `time_center` in the radio-frequency pulse's envelope.
    """
    if hasattr(rf, 'center'):
        return rf.center, np.argmin(abs(rf.t - rf.center)).item()

    # Detect the excitation peak; if i is a plateau take its center
    rf_max = np.max(np.abs(rf.signal))
    i_peak = np.where(np.abs(rf.signal) >= rf_max * 0.99999)[0]
    time_center = (rf.t[i_peak[0]] + rf.t[i_peak[-1]]) / 2
    id_center = i_peak[round((len(i_peak) - 1) / 2)]

    return time_center, id_center



import math
from types import SimpleNamespace
from typing import Tuple, Union
from warnings import warn

import numpy as np

from pypulseq import eps
from pypulseq.calc_rf_center import calc_rf_center
from pypulseq.make_trapezoid import make_trapezoid
from pypulseq.opts import Opts
from pypulseq.supported_labels_rf_use import get_supported_rf_uses
from pypulseq.utils.tracing import trace, trace_enabled


def make_adiabatic_pulse(
    pulse_type: str,
    adiabaticity: int = 4,
    bandwidth: int = 40000,
    beta: float = 800.0,
    delay: float = 0.0,
    duration: float = 10e-3,
    dwell: Union[float, None] = None,
    freq_offset: float = 0.0,
    max_grad: Union[float, None] = None,
    max_slew: Union[float, None] = None,
    n_fac: int = 40,
    mu: float = 4.9,
    phase_offset: float = 0.0,
    return_gz: bool = False,
    slice_thickness: float = 0.0,
    system: Union[Opts, None] = None,
    use: str = 'inversion',
    freq_ppm: float = 0.0,
    phase_ppm: float = 0.0,
) -> Union[
    SimpleNamespace,
    Tuple[SimpleNamespace, SimpleNamespace, SimpleNamespace],
]:
    """
    Make an adiabatic inversion pulse.

    Note: some parameters only affect certain pulse types and are ignored for other; e.g. bandwidth is ignored if
    type='hypsec'.

    hypsec(n=512, beta=800, mu=4.9, dur=0.012)
        Design a hyperbolic secant adiabatic pulse. `mu` * `beta` becomes the amplitude of the frequency sweep.

        Args:
            - n (int): number of samples (should be a multiple of 4).
            - beta (float): AM waveform parameter.
            - mu (float): a constant, determines amplitude of frequency sweep.
            - dur (float): pulse time (s).

    Returns
    -------
            2-element tuple containing
            - **a** (*array*): AM waveform.
            - **om** (*array*): FM waveform (radians/s).

    References
    ----------
            Baum, J., Tycko, R. and Pines, A. (1985). 'Broadband and adiabatic
            inversion of a two-level system by phase-modulated pulses'.
            Phys. Rev. A., 32:3435-3447.

    wurst(n=512, n_fac=40, bw=40000.0, dur=0.002)
        Design a WURST (wideband, uniform rate, smooth truncation) adiabatic inversion pulse

        Args:
            - n (int): number of samples (should be a multiple of 4).
            - n_fac (int): power to exponentiate to within AM term. ~20 or greater is typical.
            - bw (float): pulse bandwidth.
            - dur (float): pulse time (s).

    Returns
    -------
            2-element tuple containing
            - **a** (*array*): AM waveform.
            - **om** (*array*): FM waveform (radians/s).

    References
    ----------
            Kupce, E. and Freeman, R. (1995). 'Stretched Adiabatic Pulses for
            Broadband Spin Inversion'.
            J. Magn. Reason. Set. A., 117:246-256.

    Parameters
    ----------
    pulse_type : str
        One of 'hypsec' or 'wurst' pulse types.
    adiabaticity : int, default=4
    bandwidth : int, default=40000
        Pulse bandwidth.
    beta : float, default=800.0
        AM waveform parameter.
    delay : float, default=0
        Delay in seconds (s).
    duration : float, default=10e-3
        Pulse time (s).
    dwell : float, default=None
    freq_offset : float, default=0
    max_grad : float, default=None
        Maximum gradient strength.
    max_slew : float, default=None
        Maximum slew rate.
    mu : float, default=4.9
        Constant determining amplitude of frequency sweep.
    n_fac : int, default=40
        Power to exponentiate to within AM term. ~20 or greater is typical.
    phase_offset : float, default=0
        Phase offset.
    return_gz : bool, default=False
        Boolean flag to indicate if the slice-selective gradient has to be returned.
    slice_thickness : float, default=0
    system : Opts, default=Opts()
        System limits.
    use : str, default='inversion'
        Use of adiabatic radio-frequency pulse event.
        Must be one of 'excitation', 'refocusing', 'inversion',
        'saturation', 'preparation', 'other', 'undefined'.
    freq_ppm : float, default=0
        PPM frequency offset.
    phase_ppm : float, default=0
        PPM phase offset.

    Returns
    -------
    rf : SimpleNamespace
        Adiabatic RF pulse event.
    gz : SimpleNamespace, optional
        Slice-selective trapezoid event.
    gzr : SimpleNamespace, optional
        Slice-select rephasing trapezoid event.

    Raises
    ------
    ValueError
        If invalid pulse type is encountered.
        If invalid pulse use is encountered.
        If slice thickness is not provided but slice-selective trapezoid event is expected.
    """
    if system is None:
        system = Opts.default

    if return_gz and slice_thickness <= 0:
        raise ValueError('Slice thickness must be provided')

    valid_pulse_types = ['hypsec', 'wurst']
    if (not pulse_type) or (pulse_type not in valid_pulse_types):
        raise ValueError(f'Invalid type parameter. Must be one of {valid_pulse_types}.Passed: {pulse_type}')
    valid_rf_use_labels = get_supported_rf_uses()
    if use != '' and use not in valid_rf_use_labels:
        raise ValueError(f'Invalid use parameter. Must be one of {valid_rf_use_labels}. Passed: {use}')

    if dwell is None:
        dwell = system.rf_raster_time

    # WTC and PS - we have no idea why eps is added here. Leaving for now.
    n_raw = round(duration / dwell + eps)
    # Number of points must be divisible by 4 - requirement of individual pulse functions
    n_samples = math.floor(n_raw / 4) * 4

    if pulse_type == 'hypsec':
        amp_mod, freq_mod = _hypsec(n=n_samples, beta=beta, mu=mu, dur=duration)
    elif pulse_type == 'wurst':
        amp_mod, freq_mod = _wurst(n=n_samples, n_fac=n_fac, bw=bandwidth, dur=duration)

    phase_mod = np.cumsum(freq_mod) * dwell

    min_abs_freq_idx = np.argmin(np.abs(freq_mod))
    min_abs_freq_value = abs(freq_mod[min_abs_freq_idx])

    # Find rate of change of frequency at the center of the pulse
    if min_abs_freq_value == 0:
        phase_at_zero_freq = phase_mod[min_abs_freq_idx]
        amp_at_zero_freq = amp_mod[min_abs_freq_idx]
        rate_of_freq_change = abs(freq_mod[min_abs_freq_idx + 1] - freq_mod[min_abs_freq_idx - 1]) / (2 * dwell)
    else:  # We need to bracket the zero-crossing
        b = 1 if freq_mod[min_abs_freq_idx] * freq_mod[min_abs_freq_idx + 1] < 0 else -1
        diff_freq = freq_mod[min_abs_freq_idx + b] - freq_mod[min_abs_freq_idx]

        phase_at_zero_freq = (
            phase_mod[min_abs_freq_idx] * freq_mod[min_abs_freq_idx + b]
            - phase_mod[min_abs_freq_idx + b] * freq_mod[min_abs_freq_idx]
        ) / diff_freq

        amp_at_zero_freq = (
            amp_mod[min_abs_freq_idx] * freq_mod[min_abs_freq_idx + b]
            - amp_mod[min_abs_freq_idx + b] * freq_mod[min_abs_freq_idx]
        ) / diff_freq

        rate_of_freq_change = abs(freq_mod[min_abs_freq_idx] - freq_mod[min_abs_freq_idx + b]) / dwell

    # Adjust phase modulation and calculate amplitude
    phase_mod -= phase_at_zero_freq
    amp = np.sqrt(rate_of_freq_change * adiabaticity) / (2 * np.pi * amp_at_zero_freq)

    # Create the modulated signal
    signal = amp * amp_mod * np.exp(1j * phase_mod)

    # Adjust the number of samples if needed
    if n_samples != n_raw:
        n_pad = n_raw - n_samples
        pad_left = n_pad // 2
        pad_right = n_pad - pad_left
        signal = np.pad(signal, (pad_left, pad_right), mode='constant')
        n_samples = n_raw

    # Calculate time points
    t = (np.arange(n_samples) + 0.5) * dwell

    rf = SimpleNamespace()
    rf.type = 'rf'
    rf.signal = signal
    rf.t = t
    rf.shape_dur = n_samples * dwell
    rf.freq_offset = freq_offset
    rf.phase_offset = phase_offset
    rf.freq_ppm = freq_ppm
    rf.phase_ppm = phase_ppm
    rf.dead_time = system.rf_dead_time
    rf.ringdown_time = system.rf_ringdown_time
    rf.delay = delay
    rf.center, _ = calc_rf_center(rf)
    rf.use = use
    if rf.dead_time > rf.delay:
        warn(
            f'Specified RF delay {rf.delay * 1e6:.2f} us is less than the dead time {rf.dead_time * 1e6:.0f} us. Delay was increased to the dead time.',
            stacklevel=2,
        )
        rf.delay = rf.dead_time

    if return_gz:
        max_grad_slice_select = max_grad
        max_slew_slice_select = max_slew

        if pulse_type == 'hypsec':
            bandwidth = mu * beta / np.pi
        elif pulse_type == 'wurst':
            bandwidth = bandwidth

        center_pos, _ = calc_rf_center(rf)

        amplitude = bandwidth / slice_thickness
        area = amplitude * duration
        gz = make_trapezoid(
            channel='z',
            system=system,
            flat_time=duration,
            flat_area=area,
            max_grad=max_grad_slice_select,
            max_slew=max_slew_slice_select,
        )
        gzr = make_trapezoid(
            channel='z',
            system=system,
            area=-area * (1 - center_pos) - 0.5 * (gz.area - area),
            max_grad=max_grad_slice_select,
            max_slew=max_slew_slice_select,
        )

        if rf.delay > gz.rise_time:  # Round-up to gradient raster
            gz.delay = math.ceil((rf.delay - gz.rise_time) / system.grad_raster_time) * system.grad_raster_time

        if rf.delay < (gz.rise_time + gz.delay):
            rf.delay = gz.rise_time + gz.delay

    if trace_enabled():
        rf.trace = trace()

    if return_gz:
        return rf, gz, gzr
    else:
        return rf


"""Adiabatic Pulse Design functions.
    The below functions are originally from ssigpy/sigpy/mri/rf/adiabatic.py
    Used under the terms of the Sigpy BSD 3-clause license.

    Copyright (c) 2016, Frank Ong.
    Copyright (c) 2016, The Regents of the University of California.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in
       the documentation and/or other materials provided with the distribution.

    3. Neither the name of the copyright holder nor the names of its
       contributors may be used to endorse or promote products derived
       from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
    THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
    PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
    OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
    OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
    USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""


def _bir4(n: int, beta: float, kappa: float, theta: float, dw0: np.ndarray):
    r"""Design a BIR-4 adiabatic pulse.

    BIR-4 is equivalent to two BIR-1 pulses back-to-back.

    Args:
        n (int): number of samples (should be a multiple of 4).
        beta (float): AM waveform parameter.
        kappa (float): FM waveform parameter.
        theta (float): flip angle in radians.
        dw0: FM waveform scaling (radians/s).

    Returns
    -------
        2-element tuple containing

        - **a** (*array*): AM waveform.
        - **om** (*array*): FM waveform (radians/s).

    References
    ----------
        Staewen, R.S. et al. (1990). '3-D FLASH Imaging using a single surface
        coil and a new adiabatic pulse, BIR-4'.
        Invest. Radiology, 25:559-567.
    """
    dphi = np.pi + theta / 2

    t = np.arange(0, n) / n

    a1 = np.tanh(beta * (1 - 4 * t[: n // 4]))
    a2 = np.tanh(beta * (4 * t[n // 4 : n // 2] - 1))
    a3 = np.tanh(beta * (3 - 4 * t[n // 2 : 3 * n // 4]))
    a4 = np.tanh(beta * (4 * t[3 * n // 4 :] - 3))

    a = np.concatenate((a1, a2, a3, a4)).astype(np.complex64)
    a[n // 4 : 3 * n // 4] = a[n // 4 : 3 * n // 4] * np.exp(1j * dphi)

    om1 = dw0 * np.tan(kappa * 4 * t[: n // 4]) / np.tan(kappa)
    om2 = dw0 * np.tan(kappa * (4 * t[n // 4 : n // 2] - 2)) / np.tan(kappa)
    om3 = dw0 * np.tan(kappa * (4 * t[n // 2 : 3 * n // 4] - 2)) / np.tan(kappa)
    om4 = dw0 * np.tan(kappa * (4 * t[3 * n // 4 :] - 4)) / np.tan(kappa)

    om = np.concatenate((om1, om2, om3, om4))

    return a, om


def _hypsec(n: int = 512, beta: float = 800.0, mu: float = 4.9, dur: float = 0.012):
    r"""Design a hyperbolic secant adiabatic pulse.

    mu * beta becomes the amplitude of the frequency sweep

    Args:
        n (int): number of samples (should be a multiple of 4).
        beta (float): AM waveform parameter.
        mu (float): a constant, determines amplitude of frequency sweep.
        dur (float): pulse time (s).

    Returns
    -------
        2-element tuple containing

        - **a** (*array*): AM waveform.
        - **om** (*array*): FM waveform (radians/s).

    References
    ----------
        Baum, J., Tycko, R. and Pines, A. (1985). 'Broadband and adiabatic
        inversion of a two-level system by phase-modulated pulses'.
        Phys. Rev. A., 32:3435-3447.
    """
    t = np.arange(-n // 2, n // 2) / n * dur

    a = np.cosh(beta * t) ** -1
    om = -mu * beta * np.tanh(beta * t)

    return a, om


def _wurst(n: int = 512, n_fac: int = 40, bw: float = 40e3, dur: float = 2e-3):
    r"""Design a WURST (wideband, uniform rate, smooth truncation) adiabatic
     inversion pulse

    Args:
        n (int): number of samples (should be a multiple of 4).
        n_fac (int): power to exponentiate to within AM term. ~20 or greater is
         typical.
        bw (float): pulse bandwidth.
        dur (float): pulse time (s).


    Returns
    -------
        2-element tuple containing

        - **a** (*array*): AM waveform.
        - **om** (*array*): FM waveform (radians/s).

    References
    ----------
        Kupce, E. and Freeman, R. (1995). 'Stretched Adiabatic Pulses for
        Broadband Spin Inversion'.
        J. Magn. Reason. Set. A., 117:246-256.
    """
    t = np.arange(0, n) * dur / n

    a = 1 - np.power(np.abs(np.cos(np.pi * t / dur)), n_fac)
    om = np.linspace(-bw / 2, bw / 2, n) * 2 * np.pi

    return a, om


def _goia_wurst(
    n: int = 512,
    dur: float = 3.5e-3,
    f: float = 0.9,
    n_b1: int = 16,
    m_grad: int = 4,
    b1_max: float = 817,
    bw: float = 20000,
):
    r"""Design a GOIA (gradient offset independent adiabaticity) WURST
     inversion pulse

    Args:
        n (int): number of samples.
        dur (float): pulse duration (s).
        f (float): [0,1] gradient modulation factor
        n_b1 (int): order for B1 modulation
        m_grad (int): order for gradient modulation
        b1_max (float): maximum b1 (Hz)
        bw (float): pulse bandwidth (Hz)

    Returns
    -------
        3-element tuple containing:

        - **a** (*array*): AM waveform (Hz)
        - **om** (*array*): FM waveform (Hz)
        - **g** (*array*): normalized gradient waveform

    References
    ----------
        O. C. Andronesi, S. Ramadan, E.-M. Ratai, D. Jennings, C. E. Mountford,
        A. G. Sorenson.
        J Magn Reason, 203:283-293, 2010.

    """
    t = np.arange(0, n) * dur / n

    a = b1_max * (1 - np.abs(np.sin(np.pi / 2 * (2 * t / dur - 1))) ** n_b1)
    g = (1 - f) + f * np.abs(np.sin(np.pi / 2 * (2 * t / dur - 1))) ** m_grad
    om = np.cumsum((a**2) / g) * dur / n
    om = om - om[n // 2 + 1]
    om = g * om
    om = om / np.max(np.abs(om)) * bw / 2

    return a, om, g


def _bloch_siegert_fm(
    n: int = 512,
    dur: float = 2e-3,
    b1p: float = 20.0,
    k: float = 42.0,
    gamma: Union[float, None] = None,
):
    r"""
    U-shaped FM waveform for adiabatic Bloch-Siegert :math:`B_1^{+}` mapping
    and spatial encoding.

    Args:
        n (int): number of time points
        dur (float): duration in seconds
        b1p (float): nominal amplitude of constant AM waveform
        k (float): design parameter that affects max in-band
            perturbation
        gamma (float): gyromagnetic ratio

    Returns
    -------
        om (array): FM waveform (radians/s).

    References
    ----------
        M. M. Khalighi, B. K. Rutt, and A. B. Kerr.
        Adiabatic RF pulse design for Bloch-Siegert B1+ mapping.
        Magn Reason Med, 70(3):829-835, 2013.

        M. Jankiewicz, J. C. Gore, and W. A. Grissom.
        Improved encoding pulses for Bloch-Siegert B1+ mapping.
        J Magn Reason, 226:79-87, 2013.

    """
    # set gamma to PyPulseq default if not provided
    if gamma is None:
        gamma = 2 * np.pi * 42.576e6

    t = np.arange(1, n // 2) * dur / n

    om = gamma * b1p / np.sqrt((1 - gamma * b1p / k * t) ** -2 - 1)
    om = np.concatenate((om, om[::-1]))

    return om



from copy import copy
from types import SimpleNamespace
from typing import List, Union

import numpy as np

from pypulseq.calc_ramp import calc_ramp
from pypulseq.opts import Opts


def add_ramps(
    k: Union[list, np.ndarray, tuple],
    max_grad: int = 0,
    max_slew: int = 0,
    rf: Union[SimpleNamespace, None] = None,
    system: Union[Opts, None] = None,
    oversampling: bool = False,
) -> List[np.ndarray]:
    """
    Add segments to the trajectory to ramp to and from the given trajectory.

    Parameters
    ----------
    k : numpy.ndarray, or [numpy.ndarray, ...]
        If `k` is a single trajectory: Add a segment to `k` so `k_out` travels from 0 to `k[0]` and a segment so `k_out`
        goes from `k[-1]` back to 0 without violating the gradient and slew constraints.
        If `k` is multiple trajectoriess: add segments of the same length for each trajectory in the cell array.
    system : Opts, default=Opts()
        System limits.
    rf : SimpleNamespace, default=None
        Add a segment of zeros over the ramp times to an RF shape.
    max_grad : int, default=0
        Maximum gradient amplitude.
    max_slew : int, default=0
        Maximum slew rate.
    oversampling : bool, default=False
        Boolean flag to indicate if gradient is oversampled by a factor of 2.

    Returns
    -------
    result : [numpy.ndarray, ...]
        List of ramped up and ramped down k-space trajectories from `k`.

    Raises
    ------
    ValueError
        If `k` is not list, np.ndarray or tuple
    RuntimeError
        If gradient ramps fail to be calculated
    """
    if system is None:
        system = Opts.default

    if not isinstance(k, (list, np.ndarray, tuple)):
        raise ValueError(f'k has to be one of list, np.ndarray, tuple. Passed: {type(k)}')

    k_arg = copy(k)
    if max_grad > 0:
        system = copy(system)
        system.max_grad = max_grad

    if max_slew > 0:
        system = copy(system)
        system.max_slew = max_slew

    k = np.vstack(k)
    num_channels = k.shape[0]
    k = np.vstack((k, np.zeros((3 - num_channels, k.shape[1]))))  # Pad with zeros if needed

    k_up, ok1 = calc_ramp(k0=np.zeros((3, 2)), k_end=k[:, :2], system=system, oversampling=oversampling)
    k_down, ok2 = calc_ramp(k0=k[:, -2:], k_end=np.zeros((3, 2)), system=system, oversampling=oversampling)
    if not (ok1 and ok2):
        raise RuntimeError('Failed to calculate gradient ramps')

    # Add start and end points to ramps
    k_up = np.hstack((np.zeros((3, 2)), k_up))
    k_down = np.hstack((k_down, np.zeros((3, 1))))

    # Add ramps to trajectory
    k = np.hstack((k_up, k, k_down))

    result = []
    if not isinstance(k_arg, list):
        result.append(k[:num_channels])
    else:
        for i in range(num_channels):
            result.append(k[i])

    if rf is not None:
        result.append(np.concatenate((np.zeros(k_up.shape[1] * 10), rf, np.zeros(k_down.shape[1] * 10))))

    return result



import pypulseq as pp

# System settings
system = pp.Opts(
    max_grad=28,
    grad_unit='mT/m',
    max_slew=200,
    slew_unit='T/m/s',
    rf_ringdown_time=20e-6,
    rf_dead_time=100e-6,
    adc_dead_time=10e-6,
)

# System with ringdown and dead times set to 0 to introduce timing errors
system_broken = pp.Opts(
    max_grad=28,
    grad_unit='mT/m',
    max_slew=200,
    slew_unit='T/m/s',
    rf_ringdown_time=0e-6,
    rf_dead_time=0e-6,
    adc_dead_time=0e-6,
)


# Check whether there are no errors in the timing error report for the given blocks
def blocks_not_in_error_report(error_report, blocks):
    return all(error.block not in blocks for error in error_report)


# Check whether a given timing error exists in the report
def exists_in_error_report(error_report, block, event, field, error_type):
    for error in error_report:
        if error.block == block and error.event == event and error.field == field and error.error_type == error_type:
            return True
    return False


# Test whether check_timing catches all different timing errors that can occur
def test_check_timing():
    seq = pp.Sequence(system=system)

    # Add events with possible timing errors
    rf = pp.make_sinc_pulse(flip_angle=1, duration=1e-3, delay=system.rf_dead_time, system=system)
    seq.add_block(rf)  # Block 1: No error

    rf = pp.make_sinc_pulse(flip_angle=1, duration=1e-3, system=system_broken)
    seq.add_block(rf)  # Block 2: RF_DEAD_TIME, RF_RINGDOWN_TIME, BLOCK_DURATION_MISMATCH

    adc = pp.make_adc(num_samples=100, duration=1e-3, delay=system.adc_dead_time, system=system)
    seq.add_block(adc)  # Block 3: No error

    adc = pp.make_adc(num_samples=123, duration=1e-3, delay=system.adc_dead_time, system=system)
    seq.add_block(adc)  # Block 4: RASTER

    adc = pp.make_adc(num_samples=100, duration=1e-3, system=system_broken)
    seq.add_block(adc)  # Block 5: ADC_DEAD_TIME, POST_ADC_DEAD_TIME, BLOCK_DURATION_MISMATCH

    gx = pp.make_trapezoid(channel='x', area=1, duration=1, system=system)
    seq.add_block(gx)  # Block 6: No error

    gx = pp.make_trapezoid(channel='x', area=1, duration=1.00001e-3, system=system)
    seq.add_block(gx)  # Block 7: RASTER

    gx = pp.make_trapezoid(channel='x', area=1, rise_time=1e-6, flat_time=1e-3, fall_time=3e-6, system=system)
    seq.add_block(gx)  # Block 8: RASTER

    gx = pp.make_trapezoid(channel='x', area=1, duration=1e-3, delay=-1e-5, system=system)
    seq.add_block(gx)  # Block 9: NEGATIVE_DELAY

    # Check timing errors
    _, error_report = seq.check_timing()

    # Check whether the timing error report is as expected
    assert blocks_not_in_error_report(error_report, [1, 3, 6]), 'No timing errors expected on blocks 1, 3, and 6'

    assert exists_in_error_report(error_report, 2, event='rf', field='delay', error_type='RF_DEAD_TIME')
    assert exists_in_error_report(error_report, 2, event='rf', field='duration', error_type='RF_RINGDOWN_TIME')
    assert exists_in_error_report(
        error_report, 2, event='block', field='duration', error_type='BLOCK_DURATION_MISMATCH'
    )

    assert exists_in_error_report(error_report, 4, event='adc', field='dwell', error_type='RASTER')

    assert exists_in_error_report(error_report, 5, event='adc', field='delay', error_type='ADC_DEAD_TIME')
    assert exists_in_error_report(error_report, 5, event='adc', field='duration', error_type='POST_ADC_DEAD_TIME')
    assert exists_in_error_report(
        error_report, 5, event='block', field='duration', error_type='BLOCK_DURATION_MISMATCH'
    )

    assert exists_in_error_report(error_report, 7, event='block', field='duration', error_type='RASTER')
    assert exists_in_error_report(error_report, 7, event='gx', field='flat_time', error_type='RASTER')

    assert exists_in_error_report(error_report, 8, event='block', field='duration', error_type='RASTER')
    assert exists_in_error_report(error_report, 8, event='gx', field='rise_time', error_type='RASTER')
    assert exists_in_error_report(error_report, 8, event='gx', field='fall_time', error_type='RASTER')

    assert exists_in_error_report(error_report, 9, event='gx', field='delay', error_type='NEGATIVE_DELAY')

    assert len(error_report) == 13, 'Total number of timing errors was expected to be 12'



from copy import copy, deepcopy
from types import SimpleNamespace
from typing import List, Union

import numpy as np

from pypulseq import eps
from pypulseq.calc_duration import calc_duration
from pypulseq.make_arbitrary_grad import make_arbitrary_grad
from pypulseq.make_extended_trapezoid import make_extended_trapezoid
from pypulseq.make_trapezoid import make_trapezoid
from pypulseq.opts import Opts
from pypulseq.points_to_waveform import points_to_waveform
from pypulseq.utils.cumsum import cumsum
from pypulseq.utils.tracing import trace, trace_enabled


def add_gradients(
    grads: List[SimpleNamespace],
    max_grad: int = 0,
    max_slew: int = 0,
    system: Union[Opts, None] = None,
) -> SimpleNamespace:
    """
    Returns the superposition of several gradients.

    Parameters
    ----------
    grads : [SimpleNamespace, ...]
        Gradient events.
    system : Opts, default=Opts()
        System limits.
    max_grad : float, default=0
        Maximum gradient amplitude.
    max_slew : float, default=0
        Maximum slew rate.

    Returns
    -------
    grad : SimpleNamespace
        Superimposition of gradient events from `grads`.
    """
    if system is None:
        system = Opts.default

    if max_grad <= 0:
        max_grad = system.max_grad
    if max_slew <= 0:
        max_slew = system.max_slew

    if len(grads) == 0:
        raise ValueError('No gradients specified')
    if len(grads) == 1:
        # Trapezoids only require a shallow copy
        if grads[0].type == 'trap':
            grad = copy(grads[0])
        else:
            grad = deepcopy(grads[0])

        if trace_enabled():
            grad.trace = trace()
        return grad

    # First gradient defines channel
    channel = grads[0].channel

    # Check if we have a set of traps with the same timing
    if (
        all(g.type == 'trap' for g in grads)
        and all(g.rise_time == grads[0].rise_time for g in grads)
        and all(g.flat_time == grads[0].flat_time for g in grads)
        and all(g.fall_time == grads[0].fall_time for g in grads)
        and all(g.delay == grads[0].delay for g in grads)
    ):
        grad = make_trapezoid(
            grads[0].channel,
            amplitude=sum(g.amplitude for g in grads) + eps,
            rise_time=grads[0].rise_time,
            flat_time=grads[0].flat_time,
            fall_time=grads[0].fall_time,
            delay=grads[0].delay,
            system=system,
        )
        if trace_enabled():
            grad.trace = trace()
        return grad

    # Find out the general delay of all gradients and other statistics
    delays, firsts, lasts, durs, is_trap, is_arb, is_osa = [], [], [], [], [], [], []
    for ii in range(len(grads)):
        if grads[ii].channel != channel:
            raise ValueError('Cannot add gradients on different channels.')

        delays.append(grads[ii].delay)
        durs.append(calc_duration(grads[ii]))
        is_trap.append(grads[ii].type == 'trap')
        if is_trap[-1]:
            is_arb.append(False)
            is_osa.append(False)
            firsts.append(0.0)
            lasts.append(0.0)
        else:
            tt_rast = grads[ii].tt / system.grad_raster_time
            is_arb.append(np.all(np.abs(tt_rast + 0.5 - np.arange(1, len(tt_rast) + 1))) < eps)
            is_osa.append(np.all(np.abs(tt_rast - 0.5 * np.arange(1, len(tt_rast) + 1)) < eps))
            firsts.append(grads[ii].first)
            lasts.append(grads[ii].last)

    # Check if we only have arbitrary grads on irregular time samplings, optionally mixed with trapezoids
    is_etrap = np.logical_and.reduce((np.logical_not(is_trap), np.logical_not(is_arb), np.logical_not(is_osa)))
    if np.all(np.logical_or(is_trap, is_etrap)):
        # Keep shapes still rather simple
        times = []
        for ii in range(len(grads)):
            g = grads[ii]
            if g.type == 'trap':
                times.extend(cumsum(g.delay, g.rise_time, g.flat_time, g.fall_time))
            else:
                times.extend(g.delay + g.tt)

        times = np.unique(times)
        dt = times[1:] - times[:-1]
        ieps = np.flatnonzero(dt < eps)
        if np.any(ieps):
            dtx = np.array([times[0], *dt])
            dtx[ieps] = dtx[ieps] + dtx[ieps + 1]  # Assumes that no more than two too similar values can occur
            dtx = np.delete(dtx, ieps + 1)
            times = np.cumsum(dtx)

        amplitudes = np.zeros_like(times)
        for ii in range(len(grads)):
            g = grads[ii]
            if g.type == 'trap':
                if g.flat_time > 0:  # Trapezoid or triangle
                    tt = list(cumsum(g.delay, g.rise_time, g.flat_time, g.fall_time))
                    waveform = [0, g.amplitude, g.amplitude, 0]
                else:
                    tt = list(cumsum(g.delay, g.rise_time, g.fall_time))
                    waveform = [0, g.amplitude, 0]
            else:
                tt = g.delay + g.tt
                waveform = g.waveform

            # Fix rounding for the first and last time points
            i_min = np.argmin(np.abs(tt[0] - times))
            t_min = (np.abs(tt[0] - times))[i_min]
            if t_min < eps:
                tt[0] = times[i_min]
            i_min = np.argmin(np.abs(tt[-1] - times))
            t_min = (np.abs(tt[-1] - times))[i_min]
            if t_min < eps:
                tt[-1] = times[i_min]

            if abs(waveform[0]) > eps and tt[0] > eps:
                tt[0] += eps

            amplitudes += np.interp(xp=tt, fp=waveform, x=times, left=0, right=0)

        grad = make_extended_trapezoid(channel=channel, amplitudes=amplitudes, times=times, system=system)

        if trace_enabled():
            grad.trace = trace()

        return grad

    # Convert to numpy.ndarray for fancy-indexing later on
    firsts, lasts = np.array(firsts), np.array(lasts)
    common_delay = np.min(delays)
    total_duration = np.max(durs)
    durs = np.array(durs)

    # Convert everything to a regularly-sampled waveform
    waveforms = {}
    max_length = 0

    if np.any(is_osa):
        target_raster = 0.5 * system.grad_raster_time
    else:
        target_raster = system.grad_raster_time

    for ii in range(len(grads)):
        g = grads[ii]
        if g.type == 'grad':
            if is_arb[ii] or is_osa[ii]:
                if np.any(is_osa) and is_arb[ii]:  # Porting MATLAB here, maybe a bit ugly
                    # Interpolate missing samples
                    idx = np.arange(0, len(g.waveform) - 0.5 + eps, 0.5)
                    wf = g.waveform
                    interp_waveform = 0.5 * (wf[np.floor(idx).astype(int)] + wf[np.ceil(idx).astype(int)])
                    waveforms[ii] = interp_waveform
                else:
                    waveforms[ii] = g.waveform
            else:
                waveforms[ii] = points_to_waveform(
                    amplitudes=g.waveform,
                    times=g.tt,
                    grad_raster_time=target_raster,
                )
        elif g.type == 'trap':
            if g.flat_time > 0:  # Triangle or trapezoid
                times = np.array(
                    [
                        g.delay - common_delay,
                        g.delay - common_delay + g.rise_time,
                        g.delay - common_delay + g.rise_time + g.flat_time,
                        g.delay - common_delay + g.rise_time + g.flat_time + g.fall_time,
                    ]
                )
                amplitudes = np.array([0, g.amplitude, g.amplitude, 0])
            else:
                times = np.array(
                    [
                        g.delay - common_delay,
                        g.delay - common_delay + g.rise_time,
                        g.delay - common_delay + g.rise_time + g.fall_time,
                    ]
                )
                amplitudes = np.array([0, g.amplitude, 0])
            waveforms[ii] = points_to_waveform(
                amplitudes=amplitudes,
                times=times,
                grad_raster_time=target_raster,
            )
        else:
            raise ValueError('Unknown gradient type')

        if g.delay - common_delay > 0:
            # Stop for numpy.arange is not g.delay - common_delay - system.grad_raster_time like in Matlab
            # so as to include the endpoint
            waveforms[ii] = np.concatenate(
                (np.zeros(round((g.delay - common_delay) / system.grad_raster_time)), waveforms[ii])
            )

        num_points = len(waveforms[ii])
        max_length = max(num_points, max_length)

    w = np.zeros(max_length)
    for ii in range(len(grads)):
        wt = np.zeros(max_length)
        wt[0 : len(waveforms[ii])] = waveforms[ii]
        w += wt

    grad = make_arbitrary_grad(
        channel=channel,
        waveform=w,
        system=system,
        max_slew=max_slew,
        max_grad=max_grad,
        delay=common_delay,
        oversampling=np.any(is_osa),
        first=np.sum(firsts[delays == common_delay]),
        last=np.sum(lasts[durs == total_duration]),
    )
    # Fix the first and the last values
    # First is defined by the sum of firsts with the minimal delay (common_delay)
    # Last is defined by the sum of lasts with the maximum duration (total_duration == durs.max())
    grad.first = np.sum(firsts[np.array(delays) == common_delay])
    grad.last = np.sum(lasts[durs == durs.max()])

    if trace_enabled():
        grad.trace = trace()

    return grad



from types import SimpleNamespace

import numpy as np


def compress_shape(decompressed_shape: np.ndarray, force_compression: bool = False) -> SimpleNamespace:
    """
    Compress a gradient or pulse shape waveform using a run-length compression scheme on the derivative. This strategy
    encodes constant and linear waveforms with very few samples. A structure is returned with the fields:
    - num_samples - the number of samples in the uncompressed waveform
    - data - containing the compressed waveform

    See also `pypulseq.decompress_shape.py`.

    Parameters
    ----------
    decompressed_shape : numpy.ndarray
        Decompressed shape.
    force_compression: bool, default=False
        Boolean flag to indicate if compression is forced.

    Returns
    -------
    compressed_shape : SimpleNamespace
        A `SimpleNamespace` object containing the number of samples and the compressed data.
    """
    if np.any(~np.isfinite(decompressed_shape)):
        raise ValueError('compress_shape() received infinite samples.')

    if not force_compression and len(decompressed_shape) <= 4:  # Avoid compressing very short shapes
        compressed_shape = SimpleNamespace()
        compressed_shape.num_samples = len(decompressed_shape)
        compressed_shape.data = decompressed_shape
        return compressed_shape

    # Single precision floating point has ~7.25 decimal places
    quant_factor = 1e-7
    decompressed_shape_scaled = decompressed_shape / quant_factor
    datq = np.round(np.concatenate((decompressed_shape_scaled[[0]], np.diff(decompressed_shape_scaled))))
    qerr = decompressed_shape_scaled - np.cumsum(datq)
    qcor = np.concatenate(([0], np.diff(np.round(qerr))))
    datd = datq + qcor

    # RLE of datd
    starts = np.concatenate(([0], np.flatnonzero(datd[1:] != datd[:-1]) + 1))
    lengths = np.diff(np.concatenate((starts, [len(datd)])))
    values = datd[starts] * quant_factor

    # Repeat values of any run-length>1 three times: (value, value, length)
    rl_gt1 = lengths > 1
    repeats = 1 + rl_gt1 * 2
    v = np.repeat(values, repeats)

    # Calculate indices of length elements and insert length values
    inds = np.cumsum(repeats) - 1
    v[inds[rl_gt1]] = lengths[rl_gt1] - 2

    compressed_shape = SimpleNamespace()
    compressed_shape.num_samples = len(decompressed_shape)

    # Decide whether compression makes sense, otherwise store the original
    if force_compression or compressed_shape.num_samples > len(v):
        compressed_shape.data = v
    else:
        compressed_shape.data = decompressed_shape

    return compressed_shape



from types import SimpleNamespace

import numpy as np


def decompress_shape(compressed_shape: SimpleNamespace, force_decompression: bool = False) -> np.ndarray:
    """
    Decompress a gradient or pulse shape compressed with a run-length compression scheme on the derivative. The given
    shape is structure with the following fields:
    - num_samples - the number of samples in the uncompressed waveform
    - data - containing the compressed waveform

    See also `compress_shape.py`.

    Parameters
    ----------
    compressed_shape : SimpleNamespace
        Run-length encoded shape.
    force_decompression : bool, default=False

    Returns
    -------
    decompressed_shape : numpy.ndarray
        Decompressed shape.
    """
    data_pack = compressed_shape.data
    data_pack_len = len(data_pack)
    num_samples = int(compressed_shape.num_samples)

    if not force_decompression and num_samples == data_pack_len:
        # Uncompressed shape
        decompressed_shape = data_pack
        return decompressed_shape

    decompressed_shape = np.zeros(num_samples)  # Pre-allocate result matrix

    # Decompression starts here
    data_pack_diff = data_pack[1:] - data_pack[:-1]

    # When data_pack_diff == 0 the subsequent samples are equal ==> marker for repeats (run-length encoding)
    data_pack_markers = np.where(data_pack_diff == 0.0)[0]

    count_pack = 0  # Points to current compressed sample
    count_unpack = 0  # Points to current uncompressed sample

    for i in range(len(data_pack_markers)):
        # This index may have "false positives", e.g. if the value 3 repeats 3 times, then we will have 3 3 3
        next_pack = data_pack_markers[i]
        current_unpack_samples = next_pack - count_pack
        if current_unpack_samples < 0:  # Rejects false positives
            continue
        elif current_unpack_samples > 0:  # We have an unpacked block to copy
            decompressed_shape[count_unpack : count_unpack + current_unpack_samples] = data_pack[count_pack:next_pack]
            count_pack += current_unpack_samples
            count_unpack += current_unpack_samples

        # Packed/repeated section
        rep = int(data_pack[count_pack + 2] + 2)
        decompressed_shape[count_unpack : (count_unpack + rep)] = data_pack[count_pack]
        count_pack += 3
        count_unpack += rep

    # Samples left?
    if count_pack <= data_pack_len - 1:
        assert data_pack_len - count_pack == num_samples - count_unpack
        # Copy the rest of the shape, it is unpacked
        decompressed_shape[count_unpack:] = data_pack[count_pack:]

    decompressed_shape = np.cumsum(decompressed_shape)
    return decompressed_shape



# This code is a direct Python translation of the relevant functions in
# https://github.com/filip-szczepankiewicz/safe_pns_prediction/ to perform
# PNS calculations with pypulseq
#
# A small modification was made to safe_plot to plot long sequences better


# BSD 3-Clause License

# Copyright (c) 2018, Filip Szczepankiewicz and Thomas Witzel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from types import SimpleNamespace

import matplotlib.pyplot as plt
import numpy as np


def safe_example_hw():
    # function hw = safe_example_hw()
    #
    # SAFE model parameters for EXAMPLE scanner hardware (not a real scanner).
    # See comments for units.

    hw = SimpleNamespace()
    hw.name = 'MP_GPA_EXAMPLE'
    hw.checksum = '1234567890'
    hw.dependency = ''

    hw.x = SimpleNamespace()
    hw.x.tau1 = 0.20  # ms
    hw.x.tau2 = 0.03  # ms
    hw.x.tau3 = 3.00  # ms
    hw.x.a1 = 0.40
    hw.x.a2 = 0.10
    hw.x.a3 = 0.50
    hw.x.stim_limit = 30.0  # T/m/s
    hw.x.stim_thresh = 24.0  # T/m/s
    hw.x.g_scale = 0.35  # 1

    hw.y = SimpleNamespace()
    hw.y.tau1 = 1.50  # ms
    hw.y.tau2 = 2.50  # ms
    hw.y.tau3 = 0.15  # ms
    hw.y.a1 = 0.55
    hw.y.a2 = 0.15
    hw.y.a3 = 0.30
    hw.y.stim_limit = 15.0  # T/m/s
    hw.y.stim_thresh = 12.0  # T/m/s
    hw.y.g_scale = 0.31  # 1

    hw.z = SimpleNamespace()
    hw.z.tau1 = 2.00  # ms
    hw.z.tau2 = 0.12  # ms
    hw.z.tau3 = 1.00  # ms
    hw.z.a1 = 0.42
    hw.z.a2 = 0.40
    hw.z.a3 = 0.18
    hw.z.stim_limit = 25.0  # T/m/s
    hw.z.stim_thresh = 20.0  # T/m/s
    hw.z.g_scale = 0.25  # 1
    return hw


def safe_example_gwf():
    # function function [gwf, rf, dt] = safe_example_gwf()
    # Waveform with some frequency matching by Filip Szczepankiewicz.
    #
    # Waveform was optimized in the NOW framework by Jens Sjölund et al.
    # https://github.com/jsjol/NOW
    #
    # Optimization was Maxwell-compensated to remove effects of concomitant
    # gradients.
    # https://arxiv.org/ftp/arxiv/papers/1903/1903.03357.pdf

    ## STE
    dt = 1e-3  # ms

    # T/m
    gwf = 0.08 * np.array(
        [
            [0, 0, 0],
            [-0.2005, 0.9334, 0.3029],
            [-0.2050, 0.9324, 0.3031],
            [-0.2146, 0.9302, 0.3032],
            [-0.2313, 0.9263, 0.3030],
            [-0.2589, 0.9193, 0.3019],
            [-0.3059, 0.9060, 0.2980],
            [-0.3892, 0.8767, 0.2883],
            [-0.3850, 0.7147, 0.3234],
            [-0.3687, 0.5255, 0.3653],
            [-0.3509, 0.3241, 0.4070],
            [-0.3323, 0.1166, 0.4457],
            [-0.3136, -0.0906, 0.4783],
            [-0.2956, -0.2913, 0.5019],
            [-0.2790, -0.4793, 0.5139],
            [-0.2642, -0.6491, 0.5118],
            [-0.2518, -0.7957, 0.4939],
            [-0.2350, -0.8722, 0.4329],
            [-0.2187, -0.9111, 0.3541],
            [-0.2063, -0.9409, 0.2747],
            [-0.1977, -0.9627, 0.1933],
            [-0.1938, -0.9768, 0.1080],
            [-0.1967, -0.9820, 0.0159],
            [-0.2114, -0.9751, -0.0883],
            [-0.2292, -0.9219, -0.2150],
            [-0.2299, -0.8091, -0.3561],
            [-0.2290, -0.6748, -0.5011],
            [-0.2253, -0.5239, -0.6460],
            [-0.2178, -0.3620, -0.7868],
            [-0.2056, -0.1948, -0.9194],
            [-0.1391, -0.0473, -0.9908],
            [-0.0476, 0.0607, -0.9987],
            [0.0215, 0.1452, -0.9909],
            [0.0725, 0.2136, -0.9759],
            [0.1114, 0.2709, -0.9579],
            [0.1426, 0.3204, -0.9383],
            [0.1690, 0.3641, -0.9177],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [0, 0, 0],
            [-0.3734, -0.1768, 0.9125],
            [-0.3825, -0.2310, 0.8965],
            [-0.3919, -0.2895, 0.8752],
            [-0.4015, -0.3543, 0.8465],
            [-0.4108, -0.4290, 0.8065],
            [-0.4182, -0.5202, 0.7469],
            [-0.4178, -0.6423, 0.6451],
            [-0.3855, -0.8173, 0.4321],
            [-0.3110, -0.9418, 0.1401],
            [-0.2526, -0.9669, -0.0674],
            [-0.2100, -0.9541, -0.2213],
            [-0.1766, -0.9227, -0.3474],
            [-0.1491, -0.8788, -0.4570],
            [-0.1258, -0.8239, -0.5555],
            [-0.1056, -0.7583, -0.6459],
            [-0.0882, -0.6809, -0.7293],
            [-0.0734, -0.5900, -0.8061],
            [-0.0615, -0.4830, -0.8753],
            [-0.0533, -0.3556, -0.9349],
            [-0.0506, -0.2005, -0.9801],
            [-0.0575, -0.0019, -1.0000],
            [-0.0909, 0.2976, -0.9521],
            [-0.3027, 0.9509, -0.0860],
            [-0.2737, 0.9610, -0.0692],
            [-0.2524, 0.9675, -0.0596],
            [-0.2364, 0.9719, -0.0533],
            [-0.2245, 0.9749, -0.0490],
            [-0.2158, 0.9770, -0.0459],
            [-0.2097, 0.9785, -0.0439],
            [-0.2058, 0.9794, -0.0426],
            [-0.2039, 0.9798, -0.0420],
            [0, 0, 0],
        ]
    )

    rf = np.ones(gwf.shape[0])
    rf[40:] = -1

    return gwf, rf, dt


def safe_hw_check(hw):
    # function safe_hw_check(hw)
    #
    # Make sure that all is well with the hardware configuration.

    if (
        abs(hw.x.a1 + hw.x.a2 + hw.x.a3 - 1) > 0.001
        or abs(hw.y.a1 + hw.y.a2 + hw.y.a3 - 1) > 0.001
        or abs(hw.z.a1 + hw.z.a2 + hw.z.a3 - 1) > 0.001
    ):
        raise ValueError('Hardware specification a1+a2+a3 must be equal to 1!')

    axl = ['x', 'y', 'z']
    fnl = ['stim_limit', 'stim_thresh', 'tau1', 'tau2', 'tau3', 'a1', 'a2', 'a3', 'g_scale']

    for axn in axl:
        if not hasattr(hw, axn):
            raise ValueError(f"'{axn}' missing in hardware specification")

        hw_ax = getattr(hw, axn)
        for par in fnl:
            if not hasattr(hw_ax, par):
                raise ValueError(f"'{axn}.{par}' missing in hardware specification")


def safe_longest_time_const(hw):
    # function ltau = safe_longest_time_const(hw)
    # Get the longest time constant. Can be used to estimate the size of zero
    # padding.

    return max([hw.x.tau1, hw.x.tau2, hw.x.tau3, hw.y.tau1, hw.y.tau2, hw.y.tau3, hw.z.tau1, hw.z.tau2, hw.z.tau3])


def safe_pns_model(dgdt, dt, hw):
    # function stim = safe_pns_model(dgdt, dt, hw)
    #
    # dgdt (nx3) is in T/m/s
    # dt   (1x1) is in s
    # All time coefficients (a1 and tau1 etc.) are in ms.
    #
    # This PNS model is based on the SAFE-abstract
    # SAFE-Model - A New Method for Predicting Peripheral Nerve Stimulations in MRI
    # by Franz X. Herbank and Matthias Gebhardt. Abstract No 2007.
    # Proc. Intl. Soc. Mag. Res. Med. 8, 2000, Denver, Colorado, USA
    # https://cds.ismrm.org/ismrm-2000/PDF7/2007.PDF
    #
    # The main SAFE-model was coded by Thomas Witzel @ Martinos Center,
    # MGH, HMS, Boston, MA, USA.
    #
    # The code was adapted/expanded/corrected by Filip Szczepankiewicz @ LMI
    # BWH, HMS, Boston, MA, USA, and Lund University, Sweden.

    stim1 = hw.a1 * abs(safe_tau_lowpass(dgdt, hw.tau1, dt * 1000))
    stim2 = hw.a2 * safe_tau_lowpass(abs(dgdt), hw.tau2, dt * 1000)
    stim3 = hw.a3 * abs(safe_tau_lowpass(dgdt, hw.tau3, dt * 1000))

    stim = (stim1 + stim2 + stim3) / hw.stim_limit * hw.g_scale * 100

    return stim

    # Not sure where something goes awry, probably in the lowpass filter, but
    # compared to the Siemens simulator we are exactly a factor of pi off, so
    # I'm dividing the final result by pi.
    # Note also that the final result is essentially some kind of arbitrary
    # unit. - TW

    # UPDATE 210720 - The pi factor was not quite correct. Instead, the correct
    # factor was determined by the gradient scale factor (hw.g_scale, defined
    # in the .asc file). Thanks to Maxim Zaitsev for supporting this buggfix and
    # validating that the updated code is accurate. - FSz


def safe_tau_lowpass(dgdt, tau, dt, eps=1e-16):
    # function fw = safe_tau_lowpass(dgdt, tau, dt)
    #
    # Apply a RC lowpass filter with time constant tau = RC to data with sampling
    # interval dt. NOTE tau and dt need to be in the same unit (i.e. s or ms)
    # The SAFE model abstract by Hebrank et.al. just says "Lowpass with time-constant tau",
    # so I decided to make the most simple filter possible here.
    # The RC lowpass is also appealing because its something Siemens could have
    # easily implemented on their hardware stimulation monitors, so I'm probably
    # pretty close. - TW
    #
    # UPDATE 230206 - There was a factor alpha missing on the first sample it
    # has now been corrected. Thanks to Oliver Schad for finding this error.
    # - FSz

    alpha = dt / (tau + dt)

    # Calculate number of elements in filter to reach desired accuracy (eps)
    n = min(round(np.log(eps) / np.log(1 - alpha)), dgdt.shape[0])
    filt = (1 - alpha) ** np.arange(n)

    # Implements lowpass filter using convolution to get rid of for loop in original code
    return alpha * np.convolve(dgdt, filt)[: dgdt.shape[0]]


def safe_gwf_to_pns(gwf, rf, dt, hw, do_padding=True):
    # function [pns, res] = safe_gwf_to_pns(gwf, rf, dt, hw, doPadding)
    #
    # gwf (nx3) in T/m
    # dt  (1x1) in s
    # hw  (struct) is structure that describes the hardware configuration and PNS
    # response. Example: hw = safe_example_hw().
    # doPadding adds zeropadding based on the decay time.
    #
    # This PNS model is based on the SAFE-abstract
    # SAFE-Model - A New Method for Predicting Peripheral Nerve Stimulations in MRI
    # by Franz X. Herbank and Matthias Gebhardt. Abstract No 2007.
    # Proc. Intl. Soc. Mag. Res. Med. 8, 2000, Denver, Colorado, USA
    # https://cds.ismrm.org/ismrm-2000/PDF7/2007.PDF
    #
    # The main SAFE-model was coded by Thomas Witzel @ Martinos Center,
    # MGH, HMS, Boston, MA, USA.
    #
    # The code was adapted/expanded by Filip Szczepankiewicz @ LMI
    # BWH, HMS, Boston, MA, USA.

    if do_padding:
        zpt = safe_longest_time_const(hw) * 4 / 1000  # s
        pad1 = round(zpt / 4 / dt)
        pad2 = round(zpt / 1 / dt)

        gwf = np.pad(gwf, ((pad1, pad2), (0, 0)))
        rf = np.pad(rf, (pad1, pad2))

    safe_hw_check(hw)

    dgdt = np.diff(gwf, axis=0) / dt
    pns = np.zeros(dgdt.shape)

    pns[:, 0] = safe_pns_model(dgdt[:, 0], dt, hw.x)
    pns[:, 1] = safe_pns_model(dgdt[:, 1], dt, hw.y)
    pns[:, 2] = safe_pns_model(dgdt[:, 2], dt, hw.z)

    # Export relevant parameters
    res = SimpleNamespace()
    res.pns = pns
    res.gwf = gwf
    res.rf = rf
    res.dgdt = dgdt
    res.dt = dt
    res.hw = hw

    return pns, res


def safe_plot(pns, dt=None, envelope=True, envelope_points=500):
    # function h = safe_plot(pns, dt)
    # pns is relative PNS waveform (nx3)
    # dt is time step size in seconds.

    pnsnorm = np.sqrt((pns**2).sum(axis=1))

    # FZ: Added option to plot the moving maximum of pns and pnsnorm to keep
    #     plots for long sequences intelligible
    if envelope and pns.shape[0] > envelope_points:
        n_samples = int(np.ceil(pns.shape[0] / envelope_points))
        if dt is not None:
            dt *= n_samples

        if pns.shape[0] % n_samples != 0:
            pns = np.concatenate((pns, np.zeros((n_samples - pns.shape[0] % n_samples, pns.shape[1]))))
            pnsnorm = np.concatenate((pnsnorm, np.zeros((n_samples - pnsnorm.shape[0] % n_samples))))

        pns = pns.reshape(pns.shape[0] // n_samples, n_samples, pns.shape[1])
        pns = pns.max(axis=1)
        pnsnorm = pnsnorm.reshape(pnsnorm.shape[0] // n_samples, n_samples)
        pnsnorm = pnsnorm.max(axis=1)

    if dt is None:
        ttot = 1  # au
        xlabstr = 'Time [a.u.]'
    else:
        ttot = pns.shape[0] * dt * 1000  # ms
        xlabstr = 'Time [ms]'

    t = np.linspace(0, ttot, pns.shape[0])

    plt.plot(t, pns[:, 0], 'r-', t, pns[:, 1], 'g-', t, pns[:, 2], 'b-', t, pnsnorm, 'k-')

    plt.ylim([0, 120])
    plt.xlim([min(t), max(t)])

    plt.title(f'Predicted PNS ({max(pnsnorm):0.0f}%)')

    plt.xlabel(xlabstr)
    plt.ylabel('Relative stimulation [%]')

    plt.plot([0, max(t)], [max(pnsnorm), max(pnsnorm)], 'k:')

    plt.legend(
        [
            f'X ({max(pns[:, 0]):0.0f}%)',
            f'Y ({max(pns[:, 1]):0.0f}%)',
            f'Z ({max(pns[:, 2]):0.0f}%)',
            f'nrm ({max(pnsnorm):0.0f}%)',
        ],
        loc='best',
    )


def safe_example():
    # Load an example gradient waveform
    [gwf, rf, dt] = safe_example_gwf()

    # Load response parameters for example hardware
    hw = safe_example_hw()

    # Check if hardware parameters are consistent
    safe_hw_check(hw)

    # Check if this hw is part of the library (validate hw)
    # safe_hw_verify(hw)

    # Predict PNS levels
    pns, _ = safe_gwf_to_pns(gwf, rf, dt, hw, 1)

    # Plot some results
    safe_plot(pns, dt)


if __name__ == '__main__':
    safe_example()



import subprocess
from pathlib import Path
from sys import platform
from typing import Any, List, Tuple, Union

from pypulseq import Sequence

# Dictionary containing scanner definitions, target (groups), and the scanner
# detection cache.
scanner_definitions = {}
scanner_targets = {}
scanner_cache = {}


class ScannerDefinition:
    """
    Abstract base class for scanner install implementations.
    """

    def can_install(self) -> bool:
        """
        Check whether the sequence can be installed to this scanner. E.g.
        check whether a network connection can be established, whether
        transfers work, etc.

        Returns
        -------
        bool
            True when the scanner is available, False otherwise.
        """
        return True

    # Install a sequence to this scanner
    def install(self, seq: Sequence, **kwargs: Any) -> bool:
        """
        Install a given sequence to this scanner. Ideally `can_install` should
        be checked beforehand.

        Parameters
        ----------
        seq : Sequence
            Sequence object to install.
        **kwargs : Any
            The implementation for any scanner can accept additional keyword
            arguments (e.g. sequence name).

        Returns
        -------
        bool
            True if the install was successful.
        """
        raise NotImplementedError()


def register_scanner(name: str, definition: ScannerDefinition, groups: Union[List[str], None] = None) -> None:
    """
    Adds a `ScannerDefinition` to the list of known scanner targets.

    Parameters
    ----------
    name : str
        Name of the scanner target.
    definition : ScannerDefinition
        ScannerDefinition implementing the scanner's `install` functionality.
    groups : List[str], optional
        List of target groups the scanner belongs to. The default is None.
    """
    if name in scanner_definitions:
        raise ValueError(f'A target with name `{name}` already exists.')
    if name in scanner_targets:
        raise ValueError(f'A target group with name `{name}` already exists.')

    # Register scanner
    scanner_definitions[name] = definition
    scanner_targets[name] = [name]

    # Add scanner to target groups
    if groups is not None:
        for g in groups:
            if g not in scanner_targets:
                scanner_targets[g] = [name]
            else:
                scanner_targets[g].append(name)


def detect_scanner(target: Union[str, None] = None, clear_cache: bool = False) -> Tuple[str, ScannerDefinition]:
    """
    Detects whether any known scanner is available for installing pulseq
    sequences.

    `target` can specify a specific scanner, or a group of scanners as
    defined by `register_scanner`. If None, all scanners types are tested.

    Parameters
    ----------
    target : str, optional
        Type of scanner to detect. The default is None.
    clear_cache : bool, optional
        Clear the detection cache. The default is False.

    Raises
    ------
    ValueError
        If the `target` is not known.

    Returns
    -------
    Tuple[str, ScannerDefinition]
        The name of the scanner and an instance of the corresponding
        ScannerDefinition object.
    """
    if clear_cache:
        scanner_cache.clear()

    if target in scanner_cache:
        return scanner_cache[target]

    if target is not None:
        if target not in scanner_targets:
            raise ValueError('Unknown scanner target')
        sd = {k: scanner_definitions[k] for k in scanner_targets[target]}
    else:
        sd = scanner_definitions

    for scanner, definition in sd.items():
        if definition.can_install():
            scanner_cache[scanner] = (scanner, definition)
            scanner_cache[target] = (scanner, definition)
            return scanner, definition

    return None, None


def silent_call(command: str) -> int:
    """
    Calls a system command while suppressing all output.

    Parameters
    ----------
    command : str
        Command to execute.

    Returns
    -------
    int
        Status code for the command.
    """
    return subprocess.call(command, stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)  # noqa: S603


# Implementation for Siemens scanners
class SiemensDefinition(ScannerDefinition):
    """
    Scanner implementation for Siemens Numaris scanners.

    Checks whether the scanner responds to ping at the specified `ice_ip`, and
    if so, checks whether the specified `pulseq_seq_path` exists.

    When installing, saves the sequence to file, transfers it to the ICE
    machine, and renames and moves it into the `pulseq_seq_path`.
    """

    def __init__(self, ice_ip: str, pulseq_seq_path: str):
        self.ice_ip = ice_ip
        self.pulseq_seq_path = pulseq_seq_path

    def execute(self, command: str) -> bool:
        """
        Executes a command on the ICE machine.

        Parameters
        ----------
        command : str
            Command to execute.

        Returns
        -------
        bool
            Status whether the command was successfully executed.
        """
        status = silent_call(
            f'ssh -oBatchMode=yes -oStrictHostKeyChecking=no -oHostKeyAlgorithms=+ssh-rsa root@{self.ice_ip} "{command}"'
        )
        return status == 0

    def can_install(self) -> bool:
        """
        Check whether the sequence can be installed to this scanner.

        Checks:
            1. Does the scanner respond to pings?
            2. Can remote commands be executed, and does the pulseq directory
               exist on the ICE machine?

        Returns
        -------
        bool
            True if all checks pass.
        """
        if platform == 'win32':
            # windows
            ping_command = 'ping -w 1000 -n 1'
        elif platform == 'linux' or platform == 'darwin':
            # unix-like
            ping_command = 'ping -q -n -W1 -c1'

        # Does the scanner respond to pings?
        if silent_call(f'{ping_command} {self.ice_ip}') != 0:
            return False

        # Does the pulseq seq path exist on the scanner? (i.e. is this the right scanner?)
        return self.execute(f'test -d {self.pulseq_seq_path}')

    def install(
        self,
        seq: Sequence,
        name: str = 'external',
        local_filename: str = 'external.seq',
        remove_local_file: bool = True,
    ) -> bool:
        """
        Install the sequence to this Siemens scanner.

        Steps:
            1. Save the sequence to a local file
            2. Transfer the sequence file to the ICE machine with a temporary
               filename
            3. Change permissions on the sequence file
            4. Remove any existing sequence with the specified name
            5. Rename the temporary file to the specified name
            6. Remove the local file if requested

        Parameters
        ----------
        seq : Sequence
            Sequence object to install.
        name : str, optional
            Name for the sequence (.seq is added automatically if not specified).
            The default is 'external'.
        local_filename : str, optional
            Local filename to save the sequence to.
            The default is 'external.seq'.
        remove_local_file : bool, optional
            Whether or not to remove the local sequence file that was saved to
            `local_filename` after installing.
            The default is True.

        Returns
        -------
        bool
            True if the install was successful.
        """
        # Save the sequence to a local file
        seq.write(local_filename)

        # Transfer the sequence file to the ICE machine with a temporary filename
        status = silent_call(
            f'scp -oBatchMode=yes -oStrictHostKeyChecking=no -oHostKeyAlgorithms=+ssh-rsa {local_filename} root@{self.ice_ip}:{self.pulseq_seq_path}/external_tmp.seq'
        )
        if status != 0:
            return False

        # Change permissions on the sequence file
        if not self.execute(f'chmod a+rw {self.pulseq_seq_path}/external_tmp.seq'):
            return False
        # Remove any existing sequence with the specified name
        if not self.execute(f'rm -f {self.pulseq_seq_path}/{name}.seq'):
            return False
        # Rename the temporary file to the specified name
        # TODO: Create directories in pulseq path if `name` contains subdirectories (The Pulseq 1.4.3 interpreter will support subdirectories)
        if not self.execute(f'mv {self.pulseq_seq_path}/external_tmp.seq {self.pulseq_seq_path}/{name}.seq'):
            return False

        # Remove the local file if requested
        if remove_local_file:
            Path(local_filename).unlink()

        return True


# Built-in scanner definitions

# Siemens Numaris 4, two different IPs
register_scanner(
    'siemens_n4_2', SiemensDefinition('192.168.2.2', '/opt/medcom/MriCustomer/seq/pulseq'), ['siemens', 'siemens_n4']
)
register_scanner(
    'siemens_n4_3', SiemensDefinition('192.168.2.3', '/opt/medcom/MriCustomer/seq/pulseq'), ['siemens', 'siemens_n4']
)
# Siemens Numaris X
register_scanner(
    'siemens_nx', SiemensDefinition('192.168.2.2', '/opt/medcom/MriCustomer/CustomerSeq/pulseq'), ['siemens']
)



# Native Python cumsum for 2 to 5 elements
# e.g.: cumsum(1,2,3) == (1,3,6)
def cumsum(a, b, c=None, d=None, e=None):
    if e is not None:
        s1 = a + b
        s2 = s1 + c
        s3 = s2 + d
        return (a, s1, s2, s3, s3 + e)
    elif d is not None:
        s1 = a + b
        s2 = s1 + c
        return (a, s1, s2, s2 + d)
    elif c is not None:
        s = a + b
        return (a, s, s + c)
    else:
        return (a, a + b)



import sys
import traceback

# Global variables indicating whether tracing is enabled and how deep the
# calls are traced.
_tracing: bool = False
_trace_limit: int = 1


def trace_enabled() -> bool:
    """
    Returns whether tracing where sequence events and blocks were created is
    enabled.
    """
    global _tracing
    return _tracing


def enable_trace(limit: int = 1) -> None:
    """
    Enable tracing where sequence events and blocks were created. Note that
    this can slow down sequence creation.

    Parameters
    ----------
    limit : int, optional
        Limit on the stack depth that is stored. Increase when your sequence
        is created in nested functions and you want to trace back to the
        original call.
        The default is 1.
    """
    global _tracing, _trace_limit
    _tracing = True
    _trace_limit = limit


def disable_trace() -> None:
    """
    Disabling tracing where sequence events and blocks were created.
    """
    global _tracing
    _tracing = False


def trace() -> traceback.StackSummary:
    """
    Internal function to fetch a summary of the call stack.

    Returns
    -------
    traceback.StackSummary
        Call stack summary.
    """
    f = sys._getframe().f_back.f_back  # type: ignore
    return traceback.extract_stack(f, limit=_trace_limit)


def format_trace(trace: traceback.StackSummary, indent: int = 0) -> str:
    """
    Format a stack summary into a string. Optionally adds `indent` spaces
    in front of every line.

    Parameters
    ----------
    trace : traceback.StackSummary
        Call stack summary.
    indent : int, optional
        Number of spaces to indent. The default is 0.

    Returns
    -------
    str
        Stack summary formatted into a printable string.
    """
    return '\n'.join(' ' * indent + y for x in trace.format() for y in x.splitlines())



"""Tests for the make_block_pulse.py module

Will Clarke, University of Oxford, 2023
"""

from types import SimpleNamespace

import numpy as np
import pytest
from pypulseq import make_block_pulse


def test_invalid_use_error():
    with pytest.raises(ValueError, match=r'Invalid use parameter.'):
        make_block_pulse(flip_angle=np.pi, duration=1e-3, use='foo')


def test_bandwidth_and_duration_error():
    with pytest.raises(ValueError, match=r'One of bandwidth or duration must be defined, but not both.'):
        make_block_pulse(flip_angle=np.pi, duration=1e-3, bandwidth=1000)


def test_invalid_bandwidth_and_duration_error():
    with pytest.raises(ValueError, match=r'One of bandwidth or duration must be defined and be > 0.'):
        make_block_pulse(flip_angle=np.pi, duration=-1e-3)

    with pytest.raises(ValueError, match=r'One of bandwidth or duration must be defined and be > 0.'):
        make_block_pulse(flip_angle=np.pi, bandwidth=-1e3)


def test_default_duration_warning():
    with pytest.warns(UserWarning, match=r'Using default 4 ms duration for block pulse.'):
        make_block_pulse(flip_angle=np.pi)


def test_generation_methods():
    """Test minimum input cases
    Cover:
        - Just flip_angle
        - duration
        - bandwidth
        - bandwidth + time_bw_product
    """
    # Capture expected warning for default case
    with pytest.warns(UserWarning):
        case1 = make_block_pulse(flip_angle=np.pi)

    assert isinstance(case1, SimpleNamespace)
    assert case1.shape_dur == 4e-3

    case2 = make_block_pulse(flip_angle=np.pi, duration=1e-3)
    assert isinstance(case2, SimpleNamespace)
    assert case2.shape_dur == 1e-3

    case3 = make_block_pulse(flip_angle=np.pi, bandwidth=1e3)
    assert isinstance(case3, SimpleNamespace)
    assert case3.shape_dur == 1 / (4 * 1e3)

    case4 = make_block_pulse(flip_angle=np.pi, bandwidth=1e3, time_bw_product=5)
    assert isinstance(case4, SimpleNamespace)
    assert case4.shape_dur == 5 / 1e3


def test_amp_calculation():
    # A 1 ms 180 degree pulse requires 500 Hz gamma B1
    pulse = make_block_pulse(duration=1e-3, flip_angle=np.pi)
    assert np.isclose(pulse.signal.max(), 500)

    pulse = make_block_pulse(duration=1e-3, flip_angle=np.pi / 2)
    assert np.isclose(pulse.signal.max(), 250)

    pulse = make_block_pulse(duration=2e-3, flip_angle=np.pi / 2)
    assert np.isclose(pulse.signal.max(), 125)



"""Tests for SigPy pulse generation."""

import importlib.util

import numpy as np
import pypulseq as pp
import pytest
from pypulseq.opts import Opts
from pypulseq.sigpy_pulse_opts import SigpyPulseOpts


def test_sigpy_import():
    if importlib.util.find_spec('pypulseq.make_sigpy_pulse'):
        # Attempt to import and ensure no issues
        pass
    else:
        with pytest.raises(ModuleNotFoundError, match='SigPy is not installed'):
            raise ModuleNotFoundError('SigPy is not installed.')


@pytest.mark.sigpy
def test_slr():
    import sigpy.mri.rf as sigpy_rf
    from pypulseq.make_sigpy_pulse import sigpy_n_seq

    slice_thickness = 3e-3
    flip_angle = np.pi / 2
    duration = 3e-3
    time_bw_product = 4

    system = Opts(
        max_grad=32,
        grad_unit='mT/m',
        max_slew=130,
        slew_unit='T/m/s',
        rf_ringdown_time=30e-6,
        rf_dead_time=100e-6,
    )

    pulse_cfg = SigpyPulseOpts(
        pulse_type='slr',
        ptype='st',
        ftype='ls',
        d1=0.01,
        d2=0.01,
        cancel_alpha_phs=False,
        n_bands=3,
        band_sep=20,
        phs_0_pt='None',
    )

    rfp, _, _ = sigpy_n_seq(  # type: ignore
        flip_angle=flip_angle,
        delay=100e-6,
        duration=duration,
        return_gz=True,
        slice_thickness=slice_thickness,
        system=system,
        time_bw_product=time_bw_product,
        pulse_cfg=pulse_cfg,
        plot=False,
    )

    # Check that the number of samples in the pulse is correct
    assert rfp.signal.shape[0] == pytest.approx(duration / system.rf_raster_time)

    # Check that the pulse can be added to a PyPulseq Sequence
    seq = pp.Sequence(system=system)
    seq.add_block(rfp)

    # Check that the pulse shape is correct
    pulse = (rfp.signal * system.rf_raster_time * 2 * np.pi) / flip_angle
    [a, b] = sigpy_rf.sim.abrm(
        pulse,
        np.arange(-20 * time_bw_product, 20 * time_bw_product, 40 * time_bw_product / 2000),
        True,
    )
    mag_xy = 2 * np.multiply(np.conj(a), b)
    plateau_widths = np.sum(np.abs(mag_xy) > 0.8)
    assert plateau_widths == 29


@pytest.mark.sigpy
def test_sms():
    import sigpy.mri.rf as sigpy_rf
    from pypulseq.make_sigpy_pulse import sigpy_n_seq

    slice_thickness = 3e-3
    flip_angle = np.pi / 2
    duration = 3e-3
    n_bands = 3
    time_bw_product = 4

    system = Opts(
        max_grad=32,
        grad_unit='mT/m',
        max_slew=130,
        slew_unit='T/m/s',
        rf_ringdown_time=30e-6,
        rf_dead_time=100e-6,
    )

    pulse_cfg = SigpyPulseOpts(
        pulse_type='sms',
        ptype='st',
        ftype='ls',
        d1=0.01,
        d2=0.01,
        cancel_alpha_phs=False,
        n_bands=n_bands,
        band_sep=20,
        phs_0_pt='None',
    )

    rfp, _, _ = sigpy_n_seq(  # type: ignore
        flip_angle=flip_angle,
        delay=100e-6,
        duration=duration,
        return_gz=True,
        slice_thickness=slice_thickness,
        system=system,
        time_bw_product=time_bw_product,
        pulse_cfg=pulse_cfg,
        plot=False,
    )

    # Check that the number of samples in the pulse is correct
    assert rfp.signal.shape[0] == pytest.approx(duration / system.rf_raster_time)

    # Check that the pulse can be added to a PyPulseq Sequence
    seq = pp.Sequence(system=system)
    seq.add_block(rfp)

    # Check that the pulse shape is correct
    pulse = (rfp.signal * system.rf_raster_time * 2 * np.pi) / flip_angle
    [a, b] = sigpy_rf.sim.abrm(
        pulse,
        np.arange(-20 * time_bw_product, 20 * time_bw_product, 40 * time_bw_product / 2000),
        True,
    )
    mag_xy = 2 * np.multiply(np.conj(a), b)
    plateau_widths = np.sum(np.abs(mag_xy) > 0.8)
    # if slr has 29 > 0.8, then sms with MB = n_bands
    assert (29 * n_bands) == plateau_widths



"""Simple unit tests for Sequence.plot() method."""

import math

import matplotlib.pyplot as plt
import pypulseq as pp


def create_test_sequence():
    """Create a test sequence with RF, gradients, and ADC - similar to seq5() but simpler."""
    seq = pp.Sequence()
    rf, gz, gzr = pp.make_sinc_pulse(flip_angle=math.pi / 8, duration=1e-3, slice_thickness=3e-3, return_gz=True)
    gx = pp.make_trapezoid('x', flat_area=100, flat_time=5e-3)
    adc = pp.make_adc(num_samples=128, duration=5e-3)

    seq.add_block(rf, gz)
    seq.add_block(gzr)
    seq.add_block(gx, adc)
    return seq


def test_plot_default_behavior_no_figures_provided():
    """Test that default behavior works when no figures are provided."""
    plt.close('all')
    seq = create_test_sequence()

    result = seq.plot(plot_now=False)
    fig1, axes1, fig2, axes2 = result.fig1, result.ax1, result.fig2, result.ax2

    # Should create new figures
    assert isinstance(fig1, plt.Figure)
    assert isinstance(fig2, plt.Figure)
    assert len(axes1) == 3
    assert len(axes2) == 3

    # Figures should have content (subplots)
    assert len(fig1.get_axes()) == 3
    assert len(fig2.get_axes()) == 3

    plt.close('all')


def test_plot_overlay_behavior():
    """Test overlay functionality."""
    plt.close('all')
    seq = create_test_sequence()

    # First plot
    sp1 = seq.plot(plot_now=False)
    fig1_id = id(sp1.fig1)

    # Second plot with overlay
    seq.mod_grad_axis('x', 2.0)
    sp2 = seq.plot(overlay=sp1, plot_now=False)

    # Should reuse the same figure objects
    assert id(sp2.fig1) == fig1_id
    assert id(sp2.fig2) == id(sp1.fig2)

    plt.close('all')


def test_plot_stacked_behavior():
    """Test stacked plotting."""
    plt.close('all')
    seq = create_test_sequence()

    result = seq.plot(stacked=True, plot_now=False)

    assert result.fig1 is not None
    assert result.fig2 is None  # In stacked mode, fig2 is None (or unused)
    assert len(result.ax1) == 6  # 3 for RF/ADC + 3 for Gradients

    plt.close('all')



import math
import random

import numpy as np
import pytest
from pypulseq import Opts, calc_duration, make_extended_trapezoid, make_extended_trapezoid_area
from pypulseq.utils.cumsum import cumsum

system = Opts()

test_zoo = [
    (0, 0, 1),
    (0, 0, 10),
    (0, 0, 100),
    (0, 0, 10000),
    (0, 1000, 100),
    (-1000, 1000, 100),
    (-1000, 0, 100),
    (0, 0, -1),
    (0, 0, -10),
    (0, 0, -100),
    (0, 0, -10000),
    (0, 1000, -100),
    (-1000, 1000, -100),
    (-1000, 0, -100),
    (0, system.max_grad * 0.99, 10000),
    (0, system.max_grad * 0.99, -10000),
    (0, -system.max_grad * 0.99, 1000),
    (0, -system.max_grad * 0.99, -1000),
    (system.max_grad * 0.99, 0, 100),
    (system.max_grad * 0.99, 0, -100),
    (-system.max_grad * 0.99, 0, 1),
    (-system.max_grad * 0.99, 0, -1),
    (0, 100000, 1),
    (0, 100000, -1),
    (0, -100000, 1),
    (0, -100000, -1),
    (0, 90000, 0.45),
    (0, 90000, -0.45),
    (0, -90000, 0.45),
    (0, -90000, -0.45),
    (0, 10000, 0.5 * (10000) ** 2 / (system.max_slew * 0.99)),
    (0, system.max_grad * 0.99, 0.5 * (system.max_grad * 0.99) ** 2 / (system.max_slew * 0.99)),
    (system.max_grad * 0.99, system.max_grad * 0.99, 1),
    (system.max_grad * 0.99, system.max_grad * 0.99, -1),
]


@pytest.mark.parametrize('grad_start, grad_end, area', test_zoo)
def test_make_extended_trapezoid_area(grad_start, grad_end, area):
    g, _, _ = make_extended_trapezoid_area(
        channel='x', grad_start=grad_start, grad_end=grad_end, area=area, system=system
    )

    grad_ok = all(abs(g.waveform) <= system.max_grad)
    slew_ok = all(abs(np.diff(g.waveform) / np.diff(g.tt)) <= system.max_slew)

    assert pytest.approx(g.area) == area, 'Result area is not correct'
    assert grad_ok, 'Maximum gradient strength violated'
    assert slew_ok, 'Maximum slew rate violated'


random.seed(0)
test_zoo_random = [
    (
        (random.random() - 0.5) * 2 * system.max_grad * 0.99,
        (random.random() - 0.5) * 2 * system.max_grad * 0.99,
        (random.random() - 0.5) * 10000,
    )
    for _ in range(100)
]


@pytest.mark.parametrize('grad_start, grad_end, area', test_zoo_random)
def test_make_extended_trapezoid_area_random_cases(grad_start, grad_end, area):
    g, _, _ = make_extended_trapezoid_area(
        channel='x', grad_start=grad_start, grad_end=grad_end, area=area, system=system
    )

    grad_ok = all(abs(g.waveform) <= system.max_grad)
    slew_ok = all(abs(np.diff(g.waveform) / np.diff(g.tt)) <= system.max_slew)

    assert pytest.approx(g.area) == area, 'Result area is not correct'
    assert grad_ok, 'Maximum gradient strength violated'
    assert slew_ok, 'Maximum slew rate violated'


@pytest.mark.parametrize('grad_start, grad_end, area', test_zoo_random)
def test_make_extended_trapezoid_area_convert_to_arb(grad_start, grad_end, area):
    g, _, _ = make_extended_trapezoid_area(
        channel='x', grad_start=grad_start, grad_end=grad_end, area=area, system=system
    )

    g_arb, _, _ = make_extended_trapezoid_area(
        channel='x', grad_start=grad_start, grad_end=grad_end, area=area, convert_to_arbitrary=True, system=system
    )

    grad_ok = all(abs(g.waveform) <= system.max_grad)
    slew_ok = all(abs(np.diff(g.waveform) / np.diff(g.tt)) <= system.max_slew)

    grad_arb_ok = all(abs(g_arb.waveform) <= system.max_grad)
    slew_arb_ok = all(abs(np.diff(g_arb.waveform) / np.diff(g_arb.tt)) <= system.max_slew)

    assert pytest.approx(g.area) == g_arb.area, 'Area of extended trapz and arb gradient do not match'
    assert pytest.approx(g.shape_dur) == g_arb.shape_dur, 'Duration of extended trapz and arb gradient do not match'
    assert g.tt.shape[0] <= g_arb.tt.shape[0], (
        'Extended trapezoid should have less or equal number of points than arb gradient'
    )
    assert g.waveform.shape[0] <= g_arb.waveform.shape[0], (
        'Extended trapezoid should have less or equal number of points than arb gradient'
    )
    assert grad_ok == grad_arb_ok, 'Gradient strength violation between extended trapz and arb gradient'
    assert slew_ok == slew_arb_ok, 'Slew rate violation between extended trapz and arb gradient'


random.seed(0)
test_zoo_random = [
    (
        (random.random() - 0.5) * 3 * system.max_grad,
        (random.random() - 0.5) * 3 * system.max_grad,
        (random.random() - 0.5) * 10 * system.max_grad,
    )
    for _ in range(100)
]


@pytest.mark.parametrize('grad_start, grad_end, grad_amp', test_zoo_random)
def test_make_extended_trapezoid_area_recreate(grad_start, grad_end, grad_amp):
    def _to_raster(time: float) -> float:
        return np.ceil(time / system.grad_raster_time) * system.grad_raster_time

    def _calc_ramp_time(grad_amp, slew_rate, grad_start):
        return _to_raster(abs(grad_amp - grad_start) / slew_rate)

    grad_start = np.clip(grad_start, -system.max_grad, system.max_grad)
    grad_end = np.clip(grad_end, -system.max_grad, system.max_grad)

    # Construct extended gradient based on start, intermediate, and end gradient
    # strengths, assuming maximum slew rates.

    # If grad_amp > max_grad, convert to max_grad, and keep total area
    # approximately equal
    if abs(grad_amp) > system.max_grad * 0.99:
        grad_amp_new = math.copysign(system.max_grad * 0.99, grad_amp)

        flat_time = _to_raster(
            _calc_ramp_time(grad_amp_new, system.max_slew * 0.99, grad_amp)
            * abs(grad_amp_new + grad_amp)
            / abs(grad_amp_new)
        )
        grad_amp = grad_amp_new
    else:
        flat_time = 0

    # Construct extended gradient
    if flat_time == 0:
        times = cumsum(
            0,
            _calc_ramp_time(grad_amp, system.max_slew * 0.99, grad_start),
            _calc_ramp_time(grad_amp, system.max_slew * 0.99, grad_end),
        )
        amplitudes = (grad_start, grad_amp, grad_end)
    else:
        times = cumsum(
            0,
            _calc_ramp_time(grad_amp, system.max_slew * 0.99, grad_start),
            _to_raster(flat_time),
            _calc_ramp_time(grad_amp, system.max_slew * 0.99, grad_end),
        )
        amplitudes = (grad_start, grad_amp, grad_amp, grad_end)

    g_true = make_extended_trapezoid(channel='x', amplitudes=amplitudes, times=times)

    # Recreate gradient using make_extended_trapezoid_area
    g, _, _ = make_extended_trapezoid_area(
        channel='x', grad_start=grad_start, grad_end=grad_end, area=g_true.area, system=system
    )
    grad_ok = all(abs(g.waveform) <= system.max_grad)
    slew_ok = all(abs(np.diff(g.waveform) / np.diff(g.tt)) <= system.max_slew)

    assert pytest.approx(g.area) == g_true.area, 'Result area is not correct'
    assert grad_ok, 'Maximum gradient strength violated'
    assert slew_ok, 'Maximum slew rate violated'

    # Check that new gradient duration is smaller or equal to original gradient duration
    d1 = calc_duration(g)
    d2 = calc_duration(g_true)

    assert pytest.approx(d1) == d2 or d1 < d2
